from __future__ import annotations

import asyncio
import logging
import uuid
from typing import TYPE_CHECKING, Any, Mapping, Optional

import sqlalchemy as sa
from sqlalchemy.orm import load_only

from ai.backend.common.events.dispatcher import AbstractEvent
from ai.backend.common.events.event_types.kernel.broadcast import (
    BaseKernelEvent,
    KernelCancelledBroadcastEvent,
    KernelCreatingBroadcastEvent,
    KernelPreparingBroadcastEvent,
    KernelPullingBroadcastEvent,
    KernelStartedBroadcastEvent,
    KernelTerminatedBroadcastEvent,
    KernelTerminatingBroadcastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    BaseSessionEvent,
    SchedulingBroadcastEvent,
    SessionCancelledBroadcastEvent,
    SessionEnqueuedBroadcastEvent,
    SessionFailureBroadcastEvent,
    SessionScheduledBroadcastEvent,
    SessionStartedBroadcastEvent,
    SessionSuccessBroadcastEvent,
    SessionTerminatedBroadcastEvent,
    SessionTerminatingBroadcastEvent,
)
from ai.backend.common.events.hub import WILDCARD, EventPropagator
from ai.backend.common.json import dump_json_str
from ai.backend.logging import BraceStyleAdapter

if TYPE_CHECKING:
    from aiohttp_sse import EventSourceResponse

    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class SessionEventPropagator(EventPropagator):
    """
    SessionEventPropagator handles session and kernel events for SSE streaming.
    It filters events based on user permissions and session criteria.
    """

    _id: uuid.UUID
    _queue: asyncio.Queue[Optional[AbstractEvent]]
    _closed: bool
    _response: EventSourceResponse
    _db: ExtendedAsyncSAEngine
    _filters: Mapping[str, Any]
    _closed_lock: asyncio.Lock

    def __init__(
        self,
        response: EventSourceResponse,
        db: ExtendedAsyncSAEngine,
        filters: Mapping[str, Any],
    ) -> None:
        """
        Initialize the SessionEventPropagator.

        :param response: SSE response object for streaming events
        :param db: Database connection for fetching session data
        :param filters: Event filtering criteria containing:
            - user_role: User role for permission checks
            - user_uuid: User UUID for ownership checks
            - domain_name: Domain name for domain filtering
            - group_id: Group ID for group filtering
            - session_name: Session name filter
            - session_id: Session ID filter
            - access_key: Access key for session ownership
            - scope: Event scope filter ("*", "session", "kernel")
        """
        self._id = uuid.uuid4()
        self._queue = asyncio.Queue()
        self._closed = False
        self._response = response
        self._db = db
        self._filters = filters
        self._closed_lock = asyncio.Lock()

    def id(self) -> uuid.UUID:
        """Get the unique identifier for the propagator."""
        return self._id

    async def propagate_event(self, event: AbstractEvent) -> None:
        """Propagate an event to the SSE stream if it passes filtering."""
        if self._closed:
            return

        # Get event data based on event type
        data = await self._get_event_data(event)
        if data is None:
            log.warning("Could not fetch the domain object for {!r}", event)
            return
        event_name, event_data = data

        # Apply permission and session filters
        if not await self._should_send_event(event_data):
            return

        # Build response data
        response_data = {
            "reason": event_data.get("reason", ""),
            "sessionName": event_data.get("session_name", ""),
            "ownerAccessKey": event_data.get("access_key", ""),
            "sessionId": str(event_data.get("session_id", "")),
            "exitCode": event_data.get("exit_code"),
        }

        # Add kernel-specific fields if present
        if kernel_id := event_data.get("kernel_id"):
            response_data["kernelId"] = str(kernel_id)
        if cluster_role := event_data.get("cluster_role"):
            response_data["clusterRole"] = cluster_role
        if cluster_idx := event_data.get("cluster_idx"):
            response_data["clusterIdx"] = cluster_idx

        try:
            await self._response.send(dump_json_str(response_data), event=event_name)
        except Exception as e:
            log.warning("Failed to send SSE event: {}", e)
            await self.close()

    async def _get_event_data(
        self, event: AbstractEvent
    ) -> Optional[tuple[str, Mapping[str, Any]]]:
        """Get event data from database based on event type."""
        match event:
            case SchedulingBroadcastEvent():
                return await self._fetch_session_data(event)
            case (
                KernelPreparingBroadcastEvent()
                | KernelPullingBroadcastEvent()
                | KernelCreatingBroadcastEvent()
                | KernelStartedBroadcastEvent()
                | KernelCancelledBroadcastEvent()
                | KernelTerminatingBroadcastEvent()
                | KernelTerminatedBroadcastEvent()
            ):
                return await self._fetch_kernel_data(event)
            case (
                SessionEnqueuedBroadcastEvent()
                | SessionScheduledBroadcastEvent()
                | SessionStartedBroadcastEvent()
                | SessionCancelledBroadcastEvent()
                | SessionTerminatingBroadcastEvent()
                | SessionTerminatedBroadcastEvent()
                | SessionSuccessBroadcastEvent()
                | SessionFailureBroadcastEvent()
            ):
                return await self._fetch_session_data(event)
            case _:
                log.debug("Unknown event type: {}", type(event))
                return None

    async def _fetch_kernel_data(
        self, event: BaseKernelEvent
    ) -> Optional[tuple[str, Mapping[str, Any]]]:
        """Fetch kernel data from database."""
        try:
            from ai.backend.manager.models import kernels

            async with self._db.begin_readonly(isolation_level="READ COMMITTED") as conn:
                query = (
                    sa.select([
                        kernels.c.id,
                        kernels.c.session_id,
                        kernels.c.session_name,
                        kernels.c.access_key,
                        kernels.c.cluster_role,
                        kernels.c.cluster_idx,
                        kernels.c.domain_name,
                        kernels.c.group_id,
                        kernels.c.user_uuid,
                    ])
                    .select_from(kernels)
                    .where(kernels.c.id == event.kernel_id)
                )
                result = await conn.execute(query)
                row = result.first()
                if row is None:
                    return None

                data = dict(row._mapping)
                data["kernel_id"] = data["id"]
                data["reason"] = getattr(event, "reason", "")
                data["exit_Code"] = getattr(event, "exit_code", None)
                return event.event_name(), data
        except Exception as e:
            log.warning("Failed to fetch kernel data for event {}: {}", event.kernel_id, e)
            return None

    async def _fetch_session_data(
        self, event: BaseSessionEvent | SchedulingBroadcastEvent
    ) -> Optional[tuple[str, Mapping[str, Any]]]:
        """Fetch session data from database."""
        event_name = event.event_name()
        try:
            from ai.backend.manager.models.session import SessionRow

            async with self._db.begin_readonly_session(
                isolation_level="READ COMMITTED"
            ) as db_session:
                query = (
                    sa.select(SessionRow)
                    .where(SessionRow.id == event.session_id)
                    .options(
                        load_only(
                            SessionRow.id,
                            SessionRow.name,
                            SessionRow.access_key,
                            SessionRow.domain_name,
                            SessionRow.group_id,
                            SessionRow.user_uuid,
                        )
                    )
                )
                row = await db_session.scalar(query)
                if row is None:
                    return None

                data = {
                    "session_id": row.id,
                    "session_name": row.name,
                    "access_key": row.access_key,
                    "domain_name": row.domain_name,
                    "group_id": row.group_id,
                    "user_uuid": row.user_uuid,
                }
                if isinstance(event, SchedulingBroadcastEvent):
                    data["status"] = event.status_transition
                    status_to_evname = {
                        "RUNNING": "started",
                    }
                    event_name = "session_" + status_to_evname.get(
                        event.status_transition, event.status_transition.lower()
                    )
                data["reason"] = getattr(event, "reason", "")
                data["exit_Code"] = getattr(event, "exit_code", None)
                return event_name, data
        except Exception as e:
            log.warning("Failed to fetch session data for event {}: {}", event.session_id, e)
            return None

    async def _should_send_event(self, event_data: Mapping[str, Any]) -> bool:
        """Check if event should be sent based on filters."""
        from ai.backend.manager.models import UserRole

        user_role = self._filters.get("user_role")
        user_uuid = self._filters.get("user_uuid")
        domain_name = self._filters.get("domain_name")
        group_id = self._filters.get("group_id")
        session_name = self._filters.get("session_name")
        session_id = self._filters.get("session_id")
        access_key = self._filters.get("access_key")

        # Domain filtering for USER and ADMIN roles
        if user_role in (UserRole.USER, UserRole.ADMIN):
            if event_data.get("domain_name") != domain_name:
                return False

        # User ownership filtering for USER role
        if user_role == UserRole.USER:
            if event_data.get("user_uuid") != user_uuid:
                return False

        # Group filtering
        if group_id != "*" and event_data.get("group_id") != group_id:
            return False

        # Session ID filtering (takes precedence over name/access key)
        if session_id is not None and session_id != WILDCARD:
            if event_data.get("session_id") != session_id:
                return False
        else:
            # Session name and access key filtering
            if session_name != "*":
                if not (
                    event_data.get("session_name") == session_name
                    and event_data.get("access_key") == access_key
                ):
                    return False

        return True

    async def close(self) -> None:
        """Close the propagator and stop event streaming."""
        async with self._closed_lock:
            if self._closed:
                return
            self._closed = True

        # Clear the queue
        while not self._queue.empty():
            try:
                self._queue.get_nowait()
            except asyncio.QueueEmpty:
                break



import uuid
from collections.abc import Collection

import sqlalchemy as sa
from sqlalchemy.engine import Connection
from sqlalchemy.engine.row import Row

from .models import (
    get_permission_groups_table,
    get_roles_table,
)


def insert_if_data_exists(db_conn: Connection, row_type, data: Collection) -> None:
    if data:
        db_conn.execute(sa.insert(row_type), data)


def query_role_rows_by_name(db_conn: Connection, role_names: Collection[str]) -> list[Row]:
    """
    Query role rows by their names.
    """
    roles_table = get_roles_table()
    role_query = sa.select(roles_table).where(roles_table.c.name.in_(role_names))
    return db_conn.execute(role_query).all()


def query_permission_groups_by_scope_ids(
    db_conn: Connection, scope_ids: Collection[str]
) -> list[uuid.UUID]:
    """
    Query permission groups by scope IDs.
    """
    permission_groups_table = get_permission_groups_table()
    query = sa.select(permission_groups_table.c.id).where(
        permission_groups_table.c.scope_id.in_(scope_ids)
    )
    return db_conn.scalars(query).all()



from __future__ import annotations

import asyncio
import logging
import uuid
from collections.abc import (
    Awaitable,
    Callable,
    Mapping,
    Sequence,
)
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import StrEnum
from functools import partial
from typing import (
    Any,
    Optional,
    Union,
    cast,
)

import aiotools
import async_timeout
from dateutil.tz import tzutc
from sqlalchemy.exc import DBAPIError

from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.distributed import GlobalTimer
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import (
    EventProducer,
)
from ai.backend.common.events.event_types.kernel.types import (
    KernelLifecycleEventReason,
)
from ai.backend.common.events.event_types.model_serving.anycast import (
    RouteCreatedAnycastEvent,
)
from ai.backend.common.events.event_types.schedule.anycast import (
    DoCheckPrecondEvent,
    DoScaleEvent,
    DoScheduleEvent,
    DoStartSessionEvent,
)
from ai.backend.common.events.event_types.session.anycast import (
    DoUpdateSessionStatusEvent,
    SessionCancelledAnycastEvent,
    SessionCheckingPrecondAnycastEvent,
    SessionPreparingAnycastEvent,
    SessionScheduledAnycastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SessionCancelledBroadcastEvent,
    SessionPreparingBroadcastEvent,
    SessionScheduledBroadcastEvent,
)
from ai.backend.common.json import dump_json_str
from ai.backend.common.plugin.hook import PASSED, HookResult
from ai.backend.common.types import (
    AgentId,
    AgentSelectionStrategy,
    ClusterMode,
    ResourceSlot,
    SessionId,
    SessionTypes,
    aobject,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.loader.legacy_etcd_loader import LegacyEtcdLoader
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.model_serving.types import EndpointLifecycle
from ai.backend.manager.models.kernel import USER_RESOURCE_OCCUPYING_KERNEL_STATUSES
from ai.backend.manager.types import DistributedLockFactory
from ai.backend.plugin.entrypoint import scan_entrypoints

from ..defs import SERVICE_MAX_RETRIES, START_SESSION_TIMEOUT_SEC, LockID
from ..errors.common import (
    GenericBadRequest,
    GenericForbidden,
)
from ..errors.kernel import SessionNotFound
from ..errors.resource import InstanceNotAvailable
from ..exceptions import convert_to_status_data
from ..models import (
    AgentRow,
    EndpointRow,
    KernelRow,
    RouteStatus,
    ScalingGroupOpts,
    SessionRow,
)
from ..models.utils import (
    execute_with_retry,
    retry_txn,
)
from ..registry import AgentRegistry
from ..repositories.schedule.repository import ScheduleRepository
from .predicates import (
    check_concurrency,
    check_dependencies,
    check_domain_resource_limit,
    check_group_resource_limit,
    check_keypair_resource_limit,
    check_pending_session_count_limit,
    check_pending_session_resource_limit,
    check_reserved_batch_session,
    check_user_resource_limit,
)
from .types import (
    AbstractAgentSelector,
    AbstractResourceGroupState,
    AbstractScheduler,
    AgentAllocationContext,
    DefaultResourceGroupStateStore,
    KernelAgentBinding,
    PredicateResult,
    SchedulingContext,
    T_ResourceGroupState,
)

__all__ = (
    "load_scheduler",
    "load_agent_selector",
    "SchedulerDispatcher",
)

# Memoization cache for scheduler and agent selector classes
_scheduler_class_cache: dict[str, type[AbstractScheduler]] = {}
_agent_selector_class_cache: dict[str, type[AbstractAgentSelector]] = {}

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.scheduler"))

_log_fmt: ContextVar[str] = ContextVar("_log_fmt")
_log_args: ContextVar[tuple[Any, ...]] = ContextVar("_log_args")


def _load_scheduler_class(name: str) -> type[AbstractScheduler]:
    """Load and memoize scheduler class by name."""
    if name not in _scheduler_class_cache:
        entry_prefix = "backendai_scheduler_v10"
        for entrypoint in scan_entrypoints(entry_prefix):
            if entrypoint.name == name:
                log.debug('loading scheduler plugin "{}" from {}', name, entrypoint.module)
                scheduler_cls = entrypoint.load()
                _scheduler_class_cache[name] = scheduler_cls
                return scheduler_cls
        raise ImportError("Cannot load the scheduler plugin", name)
    return _scheduler_class_cache[name]


def load_scheduler(
    name: str,
    sgroup_opts: ScalingGroupOpts,
    scheduler_config: Mapping[str, Any],
) -> AbstractScheduler:
    scheduler_cls = _load_scheduler_class(name)
    return scheduler_cls(sgroup_opts, scheduler_config)


def _load_agent_selector_class(name: str) -> type[AbstractAgentSelector]:
    """Load and memoize agent selector class by name."""
    if name not in _agent_selector_class_cache:
        entry_prefix = "backendai_agentselector_v10"
        for entrypoint in scan_entrypoints(entry_prefix):
            if entrypoint.name == name:
                log.debug('loading agent-selector plugin "{}" from {}', name, entrypoint.module)
                selector_cls = entrypoint.load()
                _agent_selector_class_cache[name] = selector_cls
                return selector_cls
        raise ImportError("Cannot load the agent-selector plugin", name)
    return _agent_selector_class_cache[name]


def load_agent_selector(
    name: str,
    sgroup_opts: ScalingGroupOpts,
    selector_config: Mapping[str, Any],
    agent_selection_resource_priority: list[str],
    legacy_etcd_loader: LegacyEtcdLoader,
) -> AbstractAgentSelector[AbstractResourceGroupState]:
    def create_agent_selector(
        selector_cls: type[AbstractAgentSelector[T_ResourceGroupState]],
    ) -> AbstractAgentSelector[T_ResourceGroupState]:
        # An extra inner function to parametrize the generic type arguments
        state_cls = selector_cls.get_state_cls()
        state_store = DefaultResourceGroupStateStore(state_cls, legacy_etcd_loader)
        return selector_cls(
            sgroup_opts,
            selector_config,
            agent_selection_resource_priority,
            state_store=state_store,
        )

    selector_cls = _load_agent_selector_class(name)
    return create_agent_selector(selector_cls)


@dataclass
class LoadSchedulerArgs:
    scheduler_name: str
    sgroup_opts: ScalingGroupOpts


@dataclass
class LoadAgentSelectorArgs:
    sgroup_opts: ScalingGroupOpts
    pending_session_id: uuid.UUID
    pending_session_type: SessionTypes


class SchedulerDispatcher(aobject):
    config_provider: ManagerConfigProvider
    registry: AgentRegistry
    etcd: AsyncEtcd
    schedule_repository: ScheduleRepository

    event_producer: EventProducer
    schedule_timer: GlobalTimer
    check_precond_timer: GlobalTimer
    session_start_timer: GlobalTimer
    scale_timer: GlobalTimer
    update_session_status_timer: GlobalTimer

    _valkey_live: ValkeyLiveClient
    _valkey_stat: ValkeyStatClient

    def __init__(
        self,
        config_provider: ManagerConfigProvider,
        etcd: AsyncEtcd,
        event_producer: EventProducer,
        lock_factory: DistributedLockFactory,
        registry: AgentRegistry,
        valkey_live: ValkeyLiveClient,
        valkey_stat: ValkeyStatClient,
        schedule_repository: ScheduleRepository,
    ) -> None:
        self.config_provider = config_provider
        self.etcd = etcd
        self.event_producer = event_producer
        self.registry = registry
        self.lock_factory = lock_factory
        self._valkey_live = valkey_live
        self._valkey_stat = valkey_stat
        self.schedule_repository = schedule_repository

    @classmethod
    async def create(
        cls,
        config_provider: ManagerConfigProvider,
        etcd: AsyncEtcd,
        event_producer: EventProducer,
        lock_factory: DistributedLockFactory,
        registry: AgentRegistry,
        valkey_live: ValkeyLiveClient,
        valkey_stat: ValkeyStatClient,
        schedule_repository: ScheduleRepository,
    ) -> SchedulerDispatcher:
        instance = cls(
            config_provider,
            etcd,
            event_producer,
            lock_factory,
            registry,
            valkey_live,
            valkey_stat,
            schedule_repository,
        )
        await instance.__ainit__()
        return instance

    async def __ainit__(self) -> None:
        self.schedule_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_SCHEDULE_TIMER, 10.0),
            self.event_producer,
            lambda: DoScheduleEvent(),
            interval=10.0,
            task_name="schedule_timer",
        )
        self.session_start_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_START_TIMER, 10.0),
            self.event_producer,
            lambda: DoStartSessionEvent(),
            interval=10.0,
            initial_delay=5.0,
            task_name="session_start_timer",
        )
        self.check_precond_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_CHECK_PRECOND_TIMER, 10.0),
            self.event_producer,
            lambda: DoCheckPrecondEvent(),
            interval=10.0,
            initial_delay=5.0,
            task_name="check_precond_timer",
        )
        self.scale_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_SCALE_TIMER, 10.0),
            self.event_producer,
            lambda: DoScaleEvent(),
            interval=10.0,
            initial_delay=7.0,
            task_name="scale_timer",
        )
        self.update_session_status_timer = GlobalTimer(
            self.lock_factory(LockID.LOCKID_SESSION_STATUS_UPDATE_TIMER, 10.0),
            self.event_producer,
            lambda: DoUpdateSessionStatusEvent(),
            interval=7.0,
            initial_delay=3.0,
            task_name="update_session_status_timer",
        )
        await self.schedule_timer.join()
        await self.check_precond_timer.join()
        await self.session_start_timer.join()
        await self.scale_timer.join()
        await self.update_session_status_timer.join()

        log.info("Session scheduler started")

    async def close(self) -> None:
        async with aiotools.TaskGroup() as tg:
            tg.create_task(self.scale_timer.leave())
            tg.create_task(self.check_precond_timer.leave())
            tg.create_task(self.session_start_timer.leave())
            tg.create_task(self.schedule_timer.leave())
            tg.create_task(self.update_session_status_timer.leave())
        await self._valkey_live.close()
        await self._valkey_stat.close()
        log.info("Session scheduler stopped")

    async def schedule(
        self,
        event_name: str,
    ) -> None:
        """
        Trigger the scheduler to scan pending sessions and mark them scheduled if they fulfill
        the scheduling requirements.

        HoL blocking issue due to indefinitely preparing sessions will be mitigated because
        they will be treated as already "scheduled" sessions and the scheduler will continue to
        work on other pending sessions.

        Session status transition: PENDING -> SCHEDULED
        """
        log.debug("schedule(): triggered")
        await self._mark_scheduler_start(ScheduleType.SCHEDULE, event_name)
        known_slot_types = await self.config_provider.legacy_etcd_config_loader.get_resource_slots()
        sched_ctx = SchedulingContext(
            registry=self.registry,
            known_slot_types=known_slot_types,
        )

        lock_lifetime = self.config_provider.config.manager.session_schedule_lock_lifetime
        try:
            # The schedule() method should be executed with a global lock
            # as its individual steps are composed of many short-lived transactions.
            async with self.lock_factory(LockID.LOCKID_SCHEDULE, lock_lifetime):
                schedulable_scaling_groups = (
                    await self.schedule_repository.get_schedulable_scaling_groups()
                )
                for sgroup_name in schedulable_scaling_groups:
                    try:
                        await self._schedule_in_sgroup(
                            sched_ctx,
                            sgroup_name,
                        )
                        await self._update_scheduler_mark(
                            ScheduleType.SCHEDULE,
                            {
                                "resource_group": sgroup_name,
                            },
                        )
                    except Exception as e:
                        log.exception("schedule({}): scheduling error!\n{}", sgroup_name, repr(e))
                await self._update_scheduler_mark(
                    ScheduleType.SCHEDULE,
                    {
                        "finish_time": datetime.now(tzutc()).isoformat(),
                    },
                )
        except DBAPIError as e:
            if getattr(e.orig, "pgcode", None) == "55P03":
                log.info(
                    "schedule(): cancelled due to advisory lock timeout; "
                    "maybe another schedule() call is still running"
                )
                raise asyncio.CancelledError()
            raise

    def _load_scheduler(self, args: LoadSchedulerArgs) -> AbstractScheduler:
        global_scheduler_opts = {}
        if self.config_provider.config.plugins.scheduler:
            global_scheduler_opts = self.config_provider.config.plugins.scheduler.get(
                args.scheduler_name, {}
            )
        scheduler_config = {**global_scheduler_opts, **args.sgroup_opts.config}

        return load_scheduler(args.scheduler_name, args.sgroup_opts, scheduler_config)

    async def _load_agent_selector(self, args: LoadAgentSelectorArgs) -> AbstractAgentSelector:
        sgroup_opts = args.sgroup_opts

        # TODO: Remove "dynamic_config after refactoring.
        dynamic_config: dict[str, Any] = {}

        match sgroup_opts.agent_selection_strategy:
            case AgentSelectionStrategy.LEGACY:
                agselector_name = "legacy"
            case AgentSelectionStrategy.ROUNDROBIN:
                agselector_name = "roundrobin"
            case AgentSelectionStrategy.CONCENTRATED:
                if (
                    sgroup_opts.enforce_spreading_endpoint_replica
                    and SessionTypes(args.pending_session_type) == SessionTypes.INFERENCE
                ):
                    endpoint_id = await self.schedule_repository.get_endpoint_for_session(
                        SessionId(args.pending_session_id)
                    )
                    if endpoint_id:
                        dynamic_config[
                            "kernel_counts_at_same_endpoint"
                        ] = await self.schedule_repository.get_kernel_count_per_agent_at_endpoint(
                            endpoint_id, USER_RESOURCE_OCCUPYING_KERNEL_STATUSES
                        )

                agselector_name = "concentrated"
            case AgentSelectionStrategy.DISPERSED:
                agselector_name = "dispersed"
            case _ as unknown:
                raise ValueError(
                    f"Unknown agent selection strategy: {unknown!r}. Possible values: {[*AgentSelectionStrategy.__members__.keys()]}"
                )

        global_agselector_opts = {}
        if self.config_provider.config.plugins.agent_selector:
            global_agselector_opts = self.config_provider.config.plugins.agent_selector.get(
                agselector_name, {}
            )
        agselector_config = {
            **global_agselector_opts,
            **sgroup_opts.agent_selector_config,
            **dynamic_config,
        }

        agent_selection_resource_priority = (
            self.config_provider.config.manager.agent_selection_resource_priority
        )

        return load_agent_selector(
            agselector_name,
            sgroup_opts,
            agselector_config,
            agent_selection_resource_priority,
            self.config_provider.legacy_etcd_config_loader,
        )

    async def _schedule_in_sgroup(
        self,
        sched_ctx: SchedulingContext,
        sgroup_name: str,
    ) -> None:
        # Part 0: Load the scheduler and the agent selector.
        (
            scheduler_name,
            sgroup_opts,
        ) = await self.schedule_repository.get_scaling_group_info(sgroup_name)
        scheduler = self._load_scheduler(LoadSchedulerArgs(scheduler_name, sgroup_opts))
        (
            existing_sessions,
            pending_sessions,
            cancelled_sessions,
        ) = await self.schedule_repository.list_managed_sessions(
            sgroup_name, scheduler.sgroup_opts.pending_timeout
        )
        await self.flush_cancelled_sessions(cancelled_sessions)
        current_priority, pending_sessions = scheduler.prioritize(pending_sessions)

        log.debug(
            "running scheduler (sgroup:{}, pending:{} at prio:{}, existing:{}, cancelled:{})",
            sgroup_name,
            len(pending_sessions),
            current_priority,
            len(existing_sessions),
            len(cancelled_sessions),
        )
        num_scheduled = 0
        while len(pending_sessions) > 0:
            # Part 1: Choose the pending session to try scheduling.

            candidate_agents = await self.schedule_repository.get_schedulable_agents_by_sgroup(
                sgroup_name
            )
            total_capacity = sum((ag.available_slots for ag in candidate_agents), ResourceSlot())
            picked_session_id = scheduler.pick_session(
                total_capacity,
                pending_sessions,
                existing_sessions,
            )
            if picked_session_id is None:
                # no session is picked.
                # continue to next sgroup.
                return
            for picked_idx, pending_sess in enumerate(pending_sessions):
                if pending_sess.id == picked_session_id:
                    break
            else:
                # no matching entry for picked session?
                raise RuntimeError("should not reach here")
            pending_sess = pending_sessions.pop(picked_idx)
            log_fmt = "schedule(s:{}, prio:{}, type:{}, name:{}, ak:{}, cluster_mode:{}): "
            agent_selector = await self._load_agent_selector(
                LoadAgentSelectorArgs(
                    sgroup_opts,
                    pending_sess.id,
                    pending_sess.session_type,
                )
            )

            log_args = (
                pending_sess.id,
                pending_sess.priority,
                pending_sess.session_type,
                pending_sess.name,
                pending_sess.access_key,
                pending_sess.cluster_mode,
            )
            _log_fmt.set(log_fmt)
            _log_args.set(log_args)
            log.debug(log_fmt + "try-scheduling", *log_args)

            # Part 2: Predicate checks with predicate hook plugins

            check_results = []
            failed_predicates = []
            passed_predicates = []
            async for attempt in retry_txn():
                with attempt:
                    check_results = await self.check_predicates(
                        sched_ctx,
                        pending_sess,
                        exc_handler=lambda _: log.exception(log_fmt + "predicate-error", *log_args),
                    )
            for predicate_name, result in check_results:
                if isinstance(result, Exception):
                    failed_predicates.append({
                        "name": predicate_name,
                        "msg": repr(result),
                    })
                    continue
                if result.passed:
                    passed_predicates.append({
                        "name": predicate_name,
                    })
                else:
                    failed_predicates.append({
                        "name": predicate_name,
                        "msg": result.message or "",
                    })

            hook_result = HookResult(status=PASSED, src_plugin=[], result=[])
            async for attempt in retry_txn():
                with attempt:
                    hook_result = await self.check_predicates_hook(sched_ctx, pending_sess)
            match hook_result.src_plugin:
                case str():
                    hook_name = hook_result.src_plugin
                case list():
                    hook_name = f"({', '.join(hook_result.src_plugin)})"
                case _:
                    hook_name = ""
            if hook_result.status == PASSED:
                if hook_result.src_plugin:
                    # Append result only when plugin exists.
                    passed_predicates.append({"name": hook_name})
            else:
                failed_predicates.append({
                    "name": hook_name,
                    "msg": hook_result.reason or "",
                })

            # Part 3: Interpret the predicate check results

            status_update_data = {
                "last_try": datetime.now(tzutc()).isoformat(),
                "failed_predicates": failed_predicates,
                "passed_predicates": passed_predicates,
            }
            if failed_predicates:
                log.debug(log_fmt + "predicate-checks-failed (temporary)", *log_args)

                await self.schedule_repository.update_session_predicate_failure(
                    sched_ctx, pending_sess, status_update_data
                )
                if pending_sess.is_private:
                    await self.event_producer.anycast_and_broadcast_event(
                        SessionCancelledAnycastEvent(
                            pending_sess.id,
                            pending_sess.creation_id,
                            reason=KernelLifecycleEventReason.PENDING_TIMEOUT,
                        ),
                        SessionCancelledBroadcastEvent(
                            pending_sess.id,
                            pending_sess.creation_id,
                            reason=KernelLifecycleEventReason.PENDING_TIMEOUT,
                        ),
                    )
                # Predicate failures are *NOT* permanent errors.
                # We need to retry the scheduling afterwards.
                continue
            else:
                await self.schedule_repository.update_session_status_data(
                    pending_sess, status_update_data
                )

            # Part 4: Assign agent(s) via the agent selector.

            schedulable_sess = (
                await self.schedule_repository.get_schedulable_session_with_kernels_and_agents(
                    pending_sess.id
                )
            )

            if schedulable_sess is None:
                log.error("schedulable_sess is None for session {}", pending_sess.id)
                continue

            try:
                match ClusterMode(schedulable_sess.cluster_mode):
                    case ClusterMode.SINGLE_NODE:
                        await self._schedule_single_node_session(
                            sched_ctx,
                            agent_selector,
                            sgroup_name,
                            candidate_agents,
                            schedulable_sess,
                            check_results,
                        )
                    case ClusterMode.MULTI_NODE:
                        await self._schedule_multi_node_session(
                            sched_ctx,
                            agent_selector,
                            sgroup_name,
                            candidate_agents,
                            schedulable_sess,
                            check_results,
                        )
                    case _:
                        log.exception(
                            f"should not reach here; unknown cluster_mode: {schedulable_sess.cluster_mode}"
                        )
                        continue
                # For complex schedulers like DRF, they may need internal state updates
                # based on the scheduling result.
                scheduler.update_allocation(schedulable_sess)
            except InstanceNotAvailable as e:
                # Proceed to the next pending session and come back later.
                log.debug(
                    "schedule({}): instance not available ({})",
                    sgroup_name,
                    e.extra_msg,
                )
                continue
            except GenericBadRequest as e:
                # Proceed to the next pending session and come back later.
                log.debug(
                    "schedule({}): bad request ({})",
                    sgroup_name,
                    e.extra_msg,
                )
                continue
            except Exception:
                # _schedule_{single,multi}_node_session() already handle general exceptions.
                # Proceed to the next pending session and come back later
                continue
            num_scheduled += 1
        if num_scheduled > 0:
            await self.event_producer.anycast_event(DoCheckPrecondEvent())

    async def _filter_agent_by_container_limit(
        self, candidate_agents: list[AgentRow]
    ) -> list[AgentRow]:
        raw_value = await self.etcd.get("config/agent/max-container-count")
        if raw_value is None:
            return candidate_agents
        max_container_count = int(raw_value)

        agent_ids = [str(ag.id) for ag in candidate_agents]
        raw_counts = await self.registry.valkey_stat.get_agent_container_counts_batch(agent_ids)

        def _check(cnt: int) -> bool:
            return max_container_count > cnt

        return [ag for ag, count in zip(candidate_agents, raw_counts) if _check(count)]

    async def _schedule_single_node_session(
        self,
        sched_ctx: SchedulingContext,
        agent_selector: AbstractAgentSelector,
        sgroup_name: str,
        candidate_agents: Sequence[AgentRow],
        sess_ctx: SessionRow,
        check_results: list[tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        """
        Finds and assigns an agent having resources enough to host the entire session.
        """
        log_fmt = _log_fmt.get("")
        log_args = _log_args.get(tuple())

        try:
            requested_architectures = set(k.architecture for k in sess_ctx.kernels)
            if len(requested_architectures) > 1:
                raise GenericBadRequest(
                    "Cannot assign multiple kernels with different architectures' single node session",
                )
            if not sess_ctx.kernels:
                raise GenericBadRequest(
                    f"The session {sess_ctx.id!r} does not have any child kernel."
                )
            requested_architecture = requested_architectures.pop()
            compatible_candidate_agents = [
                ag for ag in candidate_agents if ag.architecture == requested_architecture
            ]
            if not candidate_agents:
                raise InstanceNotAvailable(extra_msg="No agents are available for scheduling")
            if not compatible_candidate_agents:
                raise InstanceNotAvailable(
                    extra_msg=(
                        "No agents found to be compatible with the image architecture "
                        f"(image[0]: {sess_ctx.main_kernel.image}, "
                        f"arch: {requested_architecture})"
                    ),
                )
            available_candidate_agents = await self._filter_agent_by_container_limit(
                compatible_candidate_agents
            )
            if not available_candidate_agents:
                raise InstanceNotAvailable(
                    extra_msg=(
                        "No agents found to be available because all agents have reached the hard"
                        " limit of the number of containers."
                    ),
                )

            # If sess_ctx.agent_id is already set for manual assignment by superadmin,
            # skip assign_agent_for_session().
            agent = sess_ctx.main_kernel.agent_row
            agent_id: AgentId | None = None
            if agent is not None:
                agent_id = agent.id

                if agent_id is not None:
                    (
                        available_slots,
                        occupied_slots,
                    ) = await self.schedule_repository.get_agent_available_slots(agent_id)

                    for key in available_slots.keys():
                        if (
                            available_slots[key] - occupied_slots.get(key, Decimal("0"))
                            >= sess_ctx.requested_slots[key]
                        ):
                            continue
                        else:
                            raise InstanceNotAvailable(
                                extra_msg=(
                                    f"The designated agent ({agent_id}) does not have "
                                    f"the enough remaining capacity ({key}, "
                                    f"requested: {sess_ctx.requested_slots[key]}, "
                                    f"remaining: {available_slots[key] - occupied_slots[key]})."
                                ),
                            )
            else:
                # Let the agent selector decide the target agent
                cand_agent_id = await agent_selector.assign_agent_for_session(
                    compatible_candidate_agents,
                    sess_ctx,
                )
                if cand_agent_id is None:
                    raise InstanceNotAvailable(
                        extra_msg=(
                            "Could not find a contiguous resource region in any agent big"
                            f" enough to host the session (id: {sess_ctx.id}, resource group:"
                            f" {sess_ctx.scaling_group_name})"
                        ),
                    )
                agent_id = cand_agent_id

            agent_alloc_ctx = await self.schedule_repository.reserve_agent(
                sgroup_name,
                agent_id,
                sess_ctx.requested_slots,
            )
        except InstanceNotAvailable as sched_failure:
            log.debug(log_fmt + "no-available-instances", *log_args)

            async def _update_sched_failure(exc: InstanceNotAvailable) -> None:
                await self.schedule_repository._update_session_scheduling_failure(
                    sched_ctx, sess_ctx, exc.extra_msg
                )

            await execute_with_retry(partial(_update_sched_failure, sched_failure))
            raise
        except Exception as e:
            log.exception(
                log_fmt + "unexpected-error, during agent allocation",
                *log_args,
            )
            exc_data = convert_to_status_data(e, self.config_provider.config.debug.enabled)

            async def _update_generic_failure() -> None:
                await self.schedule_repository._update_session_generic_failure(
                    sched_ctx, sess_ctx, exc_data
                )

            await execute_with_retry(_update_generic_failure)
            raise

        async def _finalize_scheduled() -> None:
            await self.schedule_repository.finalize_single_node_session(
                sess_ctx.id, sgroup_name, agent_alloc_ctx
            )

        await execute_with_retry(_finalize_scheduled)
        await self.registry.event_producer.anycast_and_broadcast_event(
            SessionScheduledAnycastEvent(sess_ctx.id, sess_ctx.creation_id),
            SessionScheduledBroadcastEvent(sess_ctx.id, sess_ctx.creation_id),
        )

    async def _schedule_multi_node_session(
        self,
        sched_ctx: SchedulingContext,
        agent_selector: AbstractAgentSelector,
        sgroup_name: str,
        candidate_agents: Sequence[AgentRow],
        sess_ctx: SessionRow,
        check_results: list[tuple[str, Union[Exception, PredicateResult]]],
    ) -> None:
        """
        Finds and assigns agents having resources enough to host each kernel in the session.
        """
        log_fmt = _log_fmt.get()
        log_args = _log_args.get()

        kernel_agent_bindings: list[KernelAgentBinding] = []
        # This outer transaction is rolled back when any exception occurs inside,
        # including scheduling failures of a kernel.
        # It ensures that occupied_slots are recovered when there are partial
        # scheduling failures.
        for kernel in sess_ctx.kernels:
            kernel = cast(KernelRow, kernel)
            agent_alloc_ctx: AgentAllocationContext | None = None
            agent_id: Optional[AgentId] = None
            agent: Optional[AgentRow] = kernel.agent_row
            try:
                if agent is not None:
                    # Check the resource availability of the manually designated agent
                    (
                        available_slots,
                        occupied_slots,
                    ) = await self.schedule_repository.get_agent_available_slots(agent.id)

                    for key in available_slots.keys():
                        if (
                            available_slots[key] - occupied_slots[key]
                            >= kernel.requested_slots[key]
                        ):
                            continue
                        else:
                            raise InstanceNotAvailable(
                                extra_msg=(
                                    f"The designated agent ({agent.id}) does not have "
                                    f"the enough remaining capacity ({key}, "
                                    f"requested: {sess_ctx.requested_slots[key]}, "
                                    f"remaining: {available_slots[key] - occupied_slots[key]})."
                                ),
                            )
                    agent_id = agent.id
                else:
                    # Each kernel may have different images and different architectures
                    compatible_candidate_agents = [
                        ag for ag in candidate_agents if ag.architecture == kernel.architecture
                    ]
                    if not candidate_agents:
                        raise InstanceNotAvailable(
                            extra_msg="No agents are available for scheduling"
                        )
                    if not compatible_candidate_agents:
                        raise InstanceNotAvailable(
                            extra_msg=(
                                "No agents found to be compatible with the image architecture "
                                f"(image: {kernel.image}, "
                                f"arch: {kernel.architecture})"
                            ),
                        )
                    available_candidate_agents = await self._filter_agent_by_container_limit(
                        compatible_candidate_agents
                    )
                    if not available_candidate_agents:
                        raise InstanceNotAvailable(
                            extra_msg=(
                                "No agents found to be available because all agents have"
                                " reached the hard limit of the number of containers."
                            ),
                        )
                    # Let the agent selector decide the target agent
                    agent_id = await agent_selector.assign_agent_for_kernel(
                        available_candidate_agents,
                        kernel,
                    )
                    if agent_id is None:
                        raise InstanceNotAvailable(
                            extra_msg=(
                                "Could not find a contiguous resource region in any agent big"
                                f" enough to host a kernel in the session (id: {sess_ctx.id},"
                                f" resource group: {sess_ctx.scaling_group_name})"
                            ),
                        )
                assert agent_id is not None

                agent_alloc_ctx = await self.schedule_repository.reserve_agent(
                    sgroup_name,
                    agent_id,
                    sess_ctx.requested_slots,
                )
                candidate_agents = await self.schedule_repository.get_schedulable_agents_by_sgroup(
                    sgroup_name
                )
            except InstanceNotAvailable as sched_failure:
                log.debug(log_fmt + "no-available-instances", *log_args)

                async def _update_sched_failure(exc: InstanceNotAvailable) -> None:
                    await self.schedule_repository.update_kernel_scheduling_failure(
                        sched_ctx, sess_ctx, kernel.id, exc.extra_msg
                    )

                await execute_with_retry(partial(_update_sched_failure, sched_failure))
                raise
            except Exception as e:
                log.exception(
                    log_fmt + "unexpected-error, during agent allocation",
                    *log_args,
                )
                exc_data = convert_to_status_data(e, self.config_provider.config.debug.enabled)

                async def _update_generic_failure() -> None:
                    await self.schedule_repository.update_multinode_kernel_generic_failure(
                        sched_ctx, sess_ctx, kernel.id, exc_data
                    )

                await execute_with_retry(_update_generic_failure)
                raise
            else:
                assert agent_alloc_ctx is not None
                kernel_agent_bindings.append(KernelAgentBinding(kernel, agent_alloc_ctx, set()))

        assert len(kernel_agent_bindings) == len(sess_ctx.kernels)
        # Proceed to PREPARING only when all kernels are successfully scheduled.

        async def _finalize_scheduled() -> None:
            await self.schedule_repository.finalize_multi_node_session(
                sess_ctx.id, sgroup_name, kernel_agent_bindings
            )

        await execute_with_retry(_finalize_scheduled)
        await self.registry.event_producer.anycast_and_broadcast_event(
            SessionScheduledAnycastEvent(sess_ctx.id, sess_ctx.creation_id),
            SessionScheduledBroadcastEvent(sess_ctx.id, sess_ctx.creation_id),
        )

    async def check_precond(
        self,
        event_name: str,
    ) -> None:
        """
        Scan the scheduled sessions and perform the agent RPC calls to check and pull required images.

        This function DOES NOT transit session status.
        This function calls check-and-pull API and the API produces image pull events.
        Let event handlers transit session and kernel status from
        `ImagePullStartedEvent` and `ImagePullFinishedEvent` events.
        """
        await self._mark_scheduler_start(ScheduleType.CHECK_PRECONDITION, event_name)
        lock_lifetime = self.config_provider.config.manager.session_check_precondition_lock_lifetime
        try:
            async with self.lock_factory(LockID.LOCKID_CHECK_PRECOND, lock_lifetime):
                bindings: list[KernelAgentBinding] = []

                scheduled_sessions = await self.schedule_repository.transit_scheduled_to_preparing()
                log.debug(
                    "check_precond(): checking-precond {} session(s)", len(scheduled_sessions)
                )
                for scheduled_session in scheduled_sessions:
                    for kernel in scheduled_session.kernels:
                        bindings.append(
                            KernelAgentBinding(
                                kernel=kernel,
                                agent_alloc_ctx=AgentAllocationContext(
                                    kernel.agent, kernel.agent_addr, kernel.scaling_group
                                ),
                                allocated_host_ports=set(),
                            )
                        )
                    await self.registry.event_producer.anycast_event(
                        SessionCheckingPrecondAnycastEvent(
                            scheduled_session.id,
                            scheduled_session.creation_id,
                        ),
                    )
                # check_and_pull_images() spawns tasks through PersistentTaskGroup
                await self.registry.check_and_pull_images(bindings)

            await self._update_scheduler_mark(
                ScheduleType.CHECK_PRECONDITION,
                {
                    "finish_time": datetime.now(tzutc()).isoformat(),
                },
            )
        except DBAPIError as e:
            if getattr(e.orig, "pgcode", None) == "55P03":
                log.info(
                    "check_precond(): cancelled due to advisory lock timeout; "
                    "maybe another check_precond() call is still running"
                )
                raise asyncio.CancelledError()
            raise
        except asyncio.TimeoutError:
            log.warning("check_precond(): timeout while executing start_session()")

    async def start(
        self,
        event_name: str,
    ) -> None:
        """
        Scan the sessions ready to create and perform the agent RPC calls to create kernels.

        Session status transition: PREPARED -> CREATING
        """
        await self._mark_scheduler_start(ScheduleType.START, event_name)
        lock_lifetime = self.config_provider.config.manager.session_start_lock_lifetime
        try:
            async with self.lock_factory(LockID.LOCKID_START, lock_lifetime):
                known_slot_types = (
                    await self.config_provider.legacy_etcd_config_loader.get_resource_slots()
                )
                sched_ctx = SchedulingContext(
                    self.registry,
                    known_slot_types,
                )

                scheduled_sessions = (
                    await self.schedule_repository.mark_sessions_and_kernels_creating()
                )

                log.debug("starting(): starting {} session(s)", len(scheduled_sessions))
                async with (
                    async_timeout.timeout(delay=START_SESSION_TIMEOUT_SEC),
                    aiotools.PersistentTaskGroup() as tg,
                ):
                    for scheduled_session in scheduled_sessions:
                        await self.registry.event_producer.anycast_and_broadcast_event(
                            SessionPreparingAnycastEvent(
                                scheduled_session.id,
                                scheduled_session.creation_id,
                            ),
                            SessionPreparingBroadcastEvent(
                                scheduled_session.id,
                                scheduled_session.creation_id,
                            ),
                        )
                        tg.create_task(
                            self.start_session(
                                sched_ctx,
                                scheduled_session,
                            )
                        )

            await self._update_scheduler_mark(
                ScheduleType.START,
                {
                    "finish_time": datetime.now(tzutc()).isoformat(),
                },
            )
        except DBAPIError as e:
            if getattr(e.orig, "pgcode", None) == "55P03":
                log.info(
                    "start(): cancelled due to advisory lock timeout; "
                    "maybe another start() call is still running"
                )
                raise asyncio.CancelledError()
            raise
        except asyncio.TimeoutError:
            log.warning("start(): timeout while executing start_session()")

    async def scale_services(
        self,
        event_name: str,
    ) -> None:
        log.debug("scale_services(): triggered")
        # Altering inference sessions should only be done by invoking this method
        await self._mark_scheduler_start(ScheduleType.SCALE_SERVICES, event_name)
        await execute_with_retry(lambda: self.schedule_repository.autoscale_endpoints())
        routes_to_destroy = []
        endpoints_to_expand: dict[EndpointRow, Any] = {}
        endpoints_to_mark_terminated: set[EndpointRow] = set()
        rowcount = await self.schedule_repository.clean_zombie_routes()
        if rowcount > 0:
            log.info("Cleared {} zombie routes", rowcount)

        endpoints = await self.schedule_repository.get_endpoints_for_scaling()
        for endpoint in endpoints:
            active_routings = [
                r for r in endpoint.routings if r.status in RouteStatus.active_route_statuses()
            ]
            replicas = endpoint.replicas
            if (
                endpoint.lifecycle_stage == EndpointLifecycle.DESTROYING
                and len(active_routings) == 0
            ):
                endpoints_to_mark_terminated.add(endpoint)
                continue

            if len(active_routings) > replicas:
                # We need to scale down!
                destroy_count = len(active_routings) - replicas

                # we do not expect sessions to be spawned when the endpoint is about to be destroyed
                # so also delete routes in provisioning status

                routes_to_destroy += list(
                    sorted(
                        [
                            route
                            for route in active_routings
                            if route.status in endpoint.terminatable_route_statuses
                        ],
                        key=lambda r: r.status == RouteStatus.UNHEALTHY,
                    )
                )[:destroy_count]
                log.debug(
                    "Shrinking {} from {} to {}",
                    endpoint.name,
                    len(active_routings),
                    endpoint.replicas,
                )
            elif len(active_routings) < replicas:
                if endpoint.retries > SERVICE_MAX_RETRIES:
                    continue
                # We need to scale up!
                create_count = replicas - len(active_routings)
                endpoints_to_expand[endpoint] = create_count
                log.debug(
                    "Expanding {} from {} to {}",
                    endpoint.name,
                    len(active_routings),
                    endpoint.replicas,
                )

        ids_of_session_to_destroy = [r.session for r in routes_to_destroy]
        target_sessions_to_destroy = (
            await self.schedule_repository.get_sessions_to_destroy_for_scaling(
                ids_of_session_to_destroy
            )
        )

        already_destroyed_sessions: list[SessionId] = []
        # TODO: Update logic to not to wait for sessions to actually terminate
        for session in target_sessions_to_destroy:
            try:
                await self.registry.destroy_session(
                    session,
                    forced=True,
                    reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
                )
            except (GenericForbidden, SessionNotFound):
                # Session already terminated while leaving routing alive
                already_destroyed_sessions.append(session.id)
        await self._update_scheduler_mark(
            ScheduleType.SCALE_SERVICES,
            {
                "down": dump_json_str([str(s.id) for s in target_sessions_to_destroy]),
            },
        )
        created_routes = []
        endpoint_create_data = []
        for endpoint, expand_count in endpoints_to_expand.items():
            log.debug("Creating {} session(s) for {}", expand_count, endpoint.name)
            endpoint_create_data.append((endpoint, expand_count))
        created_routes = await self.schedule_repository.create_routing_rows(endpoint_create_data)
        for route_id in created_routes:
            await self.event_producer.anycast_event(RouteCreatedAnycastEvent(route_id))
        await self._update_scheduler_mark(
            ScheduleType.SCALE_SERVICES,
            {
                "up": dump_json_str([str(e.id) for e in endpoints_to_expand.keys()]),
                "finish_time": datetime.now(tzutc()).isoformat(),
            },
        )

        await execute_with_retry(
            lambda: self.schedule_repository.destroy_terminated_endpoints_and_routes(
                endpoints_to_mark_terminated, already_destroyed_sessions
            )
        )

        await self.schedule_repository.delete_appproxy_endpoints_readonly(
            endpoints_to_mark_terminated, self.registry
        )

    async def update_session_status(
        self,
    ) -> None:
        log.debug("update_session_status(): triggered")
        candidates = await self.registry.session_lifecycle_mgr.get_status_updatable_sessions()
        await self.registry.session_lifecycle_mgr.transit_session_status(candidates)

    async def start_session(
        self,
        sched_ctx: SchedulingContext,
        session: SessionRow,
    ) -> None:
        log_fmt = (
            "start-session(s:{0.id}, type:{0.session_type}, name:{0.name}, "
            "ak:{0.access_key}, cluster_mode:{0.cluster_mode}): "
        )
        log_args = (session,)
        log.debug(log_fmt + "try-starting", *log_args)
        try:
            assert len(session.kernels) > 0
            await self.registry.start_session(sched_ctx, session)
        except (asyncio.CancelledError, Exception) as e:
            status_data = convert_to_status_data(e, self.config_provider.config.debug.enabled)
            log.warning(log_fmt + "failed-starting", *log_args, exc_info=True)
            # TODO: instead of instantly cancelling upon exception, we could mark it as
            #       SCHEDULED and retry within some limit using status_data.

            async def _mark_session_cancelled() -> None:
                await self.schedule_repository._mark_session_cancelled(
                    sched_ctx, session, status_data
                )

            log.debug(log_fmt + "cleanup-start-failure: begin", *log_args)
            try:
                await execute_with_retry(_mark_session_cancelled)
                await self.registry.event_producer.anycast_and_broadcast_event(
                    SessionCancelledAnycastEvent(
                        session.id,
                        session.creation_id,
                        KernelLifecycleEventReason.FAILED_TO_START,
                    ),
                    SessionCancelledBroadcastEvent(
                        session.id,
                        session.creation_id,
                        KernelLifecycleEventReason.FAILED_TO_START,
                    ),
                )
                cid_map = await self.schedule_repository.get_container_info_for_destroyed_kernels(
                    session.id
                )
                destroyed_kernels = [
                    {
                        "agent": k.agent,
                        "agent_addr": k.agent_addr,
                        "id": k.id,
                        "container_id": cid_map[k.id],
                    }
                    for k in session.kernels
                ]
                await self.registry.destroy_session_lowlevel(
                    session.id,
                    destroyed_kernels,
                )
                await self.registry.recalc_resource_usage()
            except Exception as destroy_err:
                log.error(log_fmt + "cleanup-start-failure: error", *log_args, exc_info=destroy_err)
            finally:
                log.debug(log_fmt + "cleanup-start-failure: done", *log_args)
        else:
            log.info(log_fmt + "started", *log_args)

    async def flush_cancelled_sessions(self, cancelled_sessions: Sequence[SessionRow]) -> None:
        if not cancelled_sessions:
            return
        session_ids = [item.id for item in cancelled_sessions]

        await self.schedule_repository.apply_cancellation(session_ids)
        for item in cancelled_sessions:
            await self.event_producer.anycast_and_broadcast_event(
                SessionCancelledAnycastEvent(
                    item.id,
                    item.creation_id,
                    reason=KernelLifecycleEventReason.PENDING_TIMEOUT,
                ),
                SessionCancelledBroadcastEvent(
                    item.id,
                    item.creation_id,
                    reason=KernelLifecycleEventReason.PENDING_TIMEOUT,
                ),
            )

    async def check_predicates(
        self,
        sched_ctx: SchedulingContext,
        pending_sess: SessionRow,
        *,
        exc_handler: Callable[[Exception], None] | None = None,
    ) -> list[tuple[str, Union[Exception, PredicateResult]]]:
        check_results: list[tuple[str, Union[Exception, PredicateResult]]] = []
        async with self.registry.db.begin_session() as db_sess:
            predicates: list[tuple[str, Awaitable[PredicateResult]]] = [
                (
                    "reserved_time",
                    check_reserved_batch_session(db_sess, sched_ctx, pending_sess),
                ),
                ("dependencies", check_dependencies(db_sess, sched_ctx, pending_sess)),
                ("concurrency", check_concurrency(db_sess, sched_ctx, pending_sess)),
            ]
            if not pending_sess.is_private:
                predicates += [
                    (
                        "pending_session_resource_limit",
                        check_pending_session_resource_limit(db_sess, sched_ctx, pending_sess),
                    ),
                    (
                        "pending_session_count_limit",
                        check_pending_session_count_limit(db_sess, sched_ctx, pending_sess),
                    ),
                    (
                        "keypair_resource_limit",
                        check_keypair_resource_limit(db_sess, sched_ctx, pending_sess),
                    ),
                    (
                        "user_resource_limit",
                        check_user_resource_limit(db_sess, sched_ctx, pending_sess),
                    ),
                    (
                        "user_group_resource_limit",
                        check_group_resource_limit(db_sess, sched_ctx, pending_sess),
                    ),
                    (
                        "domain_resource_limit",
                        check_domain_resource_limit(db_sess, sched_ctx, pending_sess),
                    ),
                ]
            for predicate_name, check_coro in predicates:
                try:
                    check_results.append((predicate_name, await check_coro))
                except DBAPIError:
                    raise
                except Exception as e:
                    if exc_handler is not None:
                        exc_handler(e)
                    check_results.append((predicate_name, e))
        return check_results

    async def check_predicates_hook(
        self,
        sched_ctx: SchedulingContext,
        pending_sess: SessionRow,
    ) -> HookResult:
        async with self.registry.db.begin_readonly_session() as db_sess:
            return await self.registry.hook_plugin_ctx.dispatch(
                "PREDICATE",
                (
                    db_sess,
                    sched_ctx,
                    pending_sess,
                ),
            )

    def _schedule_key(self, schedule_type: ScheduleType) -> str:
        """
        Returns the Redis key for the given schedule type.
        """
        manager_id = self.config_provider.config.manager.id
        return f"manager.{manager_id}.{str(schedule_type)}"

    async def _mark_scheduler_start(
        self,
        schedule_type: ScheduleType,
        event_name: str,
    ) -> None:
        schedule_key = self._schedule_key(schedule_type)
        await self._valkey_live.replace_schedule_data(
            schedule_key,
            {
                "trigger_event": event_name,
                "execution_time": datetime.now(tzutc()).isoformat(),
            },
        )

    async def _update_scheduler_mark(
        self,
        schedule_type: ScheduleType,
        to_update: dict[str, Any],
    ) -> None:
        """
        Updates the scheduler mark for the given schedule type with the provided data.
        """
        schedule_key = self._schedule_key(schedule_type)
        await self._valkey_live.add_scheduler_metadata(
            schedule_key,
            to_update,
        )


class ScheduleType(StrEnum):
    SCHEDULE = "schedule"
    START = "start"
    CHECK_PRECONDITION = "check_precondition"
    SCALE_SERVICES = "scale_services"



from typing import override

from ai.backend.manager.actions.action import BaseAction


class ObjectStorageAction(BaseAction):
    @override
    @classmethod
    def entity_type(cls) -> str:
        return "object_storage"



"""Session validator that applies multiple validation rules."""

import logging
from collections.abc import Iterable

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationContext,
    SessionCreationSpec,
)

from .base import SessionValidatorRule

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class SessionValidator:
    """
    Validator that applies multiple validation rules for session creation.
    """

    _rules: Iterable[SessionValidatorRule]

    def __init__(self, rules: Iterable[SessionValidatorRule]) -> None:
        self._rules = rules

    def validate(
        self,
        spec: SessionCreationSpec,
        context: SessionCreationContext,
    ) -> None:
        """
        Validate the session creation specification with all rules.

        Args:
            spec: Session creation specification
            context: Pre-fetched context with all required data (includes allowed_scaling_groups)

        Raises:
            InvalidAPIParameters or QuotaExceeded: If any validation fails
        """
        for rule in self._rules:
            log.debug(f"Applying validation rule: {rule.name()}")
            rule.validate(spec, context, context.allowed_scaling_groups)



import logging
import zlib
from collections.abc import Collection, Sequence
from typing import cast

import sqlalchemy as sa
from sqlalchemy.orm import contains_eager

from ai.backend.common import msgpack
from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.metrics.metric import LayerType
from ai.backend.common.types import AgentId
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.agent.modifier import AgentStatusModifier
from ai.backend.manager.data.agent.types import (
    AgentData,
    AgentDataExtended,
    AgentHeartbeatUpsert,
    UpsertResult,
)
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.models.agent import AgentRow
from ai.backend.manager.models.kernel import KernelRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.agent.cache_source.cache_source import AgentCacheSource
from ai.backend.manager.repositories.agent.db_source.db_source import AgentDBSource
from ai.backend.manager.repositories.resource_preset.utils import suppress_with_log

from .query import QueryCondition, QueryOrder

# Layer-specific decorator for agent repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.AGENT)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class AgentRepository:
    _db_source: AgentDBSource
    _cache_source: AgentCacheSource
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        valkey_image: ValkeyImageClient,
        valkey_live: ValkeyLiveClient,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._db_source = AgentDBSource(db)
        self._cache_source = AgentCacheSource(valkey_image, valkey_live)
        self._config_provider = config_provider

    @repository_decorator()
    async def get_by_id(self, agent_id: AgentId) -> AgentData:
        return await self._db_source.get_by_id(agent_id)

    @repository_decorator()
    async def add_agent_to_images(self, agent_id: AgentId, images) -> None:
        images = msgpack.unpackb(zlib.decompress(images))
        image_canonicals = set(img_info[0] for img_info in images)
        with suppress_with_log(
            [Exception], message=f"Failed to cache agent: {agent_id} to images: {image_canonicals}"
        ):
            await self._cache_source.set_agent_to_images(agent_id, list(image_canonicals))

    @repository_decorator()
    async def sync_agent_heartbeat(
        self,
        agent_id: AgentId,
        upsert_data: AgentHeartbeatUpsert,
    ) -> UpsertResult:
        with suppress_with_log(
            [Exception], message=f"Failed to update last seen for agent: {agent_id}"
        ):
            await self._cache_source.update_agent_last_seen(
                agent_id, upsert_data.heartbeat_received
            )

        upsert_result = await self._db_source.upsert_agent_with_state(upsert_data)
        if upsert_result.need_resource_slot_update:
            await self._config_provider.legacy_etcd_config_loader.update_resource_slots(
                upsert_data.resource_info.slot_key_and_units
            )

        return upsert_result

    @repository_decorator()
    async def cleanup_agent_on_exit(self, agent_id: AgentId, modifier: AgentStatusModifier) -> None:
        with suppress_with_log(
            [Exception], message=f"Failed to remove last seen for agent: {agent_id}"
        ):
            await self._cache_source.remove_agent_last_seen(agent_id)

        await self._db_source.update_agent_status_exit(agent_id, modifier)

        with suppress_with_log(
            [Exception], message=f"Failed to remove agent: {agent_id} from all images"
        ):
            await self._cache_source.remove_agent_from_all_images(agent_id)

    @repository_decorator()
    async def update_agent_status(self, agent_id: AgentId, modifier: AgentStatusModifier) -> None:
        await self._db_source.update_agent_status(agent_id, modifier)

    @repository_decorator()
    async def remove_agent_from_images(
        self, agent_id: AgentId, image_canonicals: list[str]
    ) -> None:
        with suppress_with_log(
            [Exception],
            message=f"Failed to remove agent: {agent_id} from images: {image_canonicals}",
        ):
            await self._cache_source.remove_agent_from_images(agent_id, image_canonicals)

    @repository_decorator()
    async def list_data(
        self,
        conditions: Sequence[QueryCondition],
        order_by: Sequence[QueryOrder] = tuple(),
    ) -> list[AgentData]:
        stmt: sa.sql.Select = sa.select(AgentRow)
        for cond in conditions:
            stmt = cond(stmt)

        if order_by:
            stmt = stmt.order_by(*order_by)

        async with self._db_source._db.begin_readonly_session() as db_session:
            result = await db_session.scalars(stmt)
            agent_rows = cast(list[AgentRow], result.unique().all())
            return [agent_row.to_data() for agent_row in agent_rows]

    @repository_decorator()
    async def list_extended_data(
        self,
        conditions: Sequence[QueryCondition],
        order_by: Sequence[QueryOrder] = tuple(),
        *,
        kernel_statuses: Collection[KernelStatus] = (KernelStatus.RUNNING,),
    ) -> list[AgentDataExtended]:
        stmt: sa.sql.Select = (
            sa.select(AgentRow)
            .select_from(
                sa.join(
                    AgentRow,
                    KernelRow,
                    sa.and_(AgentRow.id == KernelRow.agent, KernelRow.status.in_(kernel_statuses)),
                    isouter=True,
                )
            )
            .options(
                contains_eager(AgentRow.kernels),
            )
        )
        for cond in conditions:
            stmt = cond(stmt)

        if order_by:
            stmt = stmt.order_by(*order_by)

        known_slot_types = (
            await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        )
        async with self._db_source._db.begin_readonly_session() as db_session:
            result = await db_session.scalars(stmt)
            agent_rows = cast(list[AgentRow], result.unique().all())
            return [agent_row.to_extended_data(known_slot_types) for agent_row in agent_rows]



"""Base action for deployment service."""

from typing import override

from ai.backend.manager.actions.action import BaseAction


class DeploymentBaseAction(BaseAction):
    """Base action for deployment operations."""

    @override
    @classmethod
    def entity_type(cls) -> str:
        return "deployment"



import copy
import logging
import uuid
from datetime import datetime
from typing import Optional, Sequence, cast
from uuid import UUID

import msgpack
import sqlalchemy as sa

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import (
    InvalidAPIParameters,
)
from ai.backend.common.metrics.metric import LayerType
from ai.backend.common.utils import nmget
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.group.types import GroupCreator, GroupData, GroupModifier
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.resource import ProjectNotFound
from ai.backend.manager.models.domain import domains
from ai.backend.manager.models.group import GroupRow, association_groups_users, groups
from ai.backend.manager.models.kernel import LIVE_STATUS, RESOURCE_USAGE_KERNEL_STATUSES, kernels
from ai.backend.manager.models.resource_policy import keypair_resource_policies
from ai.backend.manager.models.resource_usage import fetch_resource_usage
from ai.backend.manager.models.user import UserRole, users
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, SASession

from ..permission_controller.role_manager import RoleManager

# Layer-specific decorator for group repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.GROUP)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class GroupRepository:
    _db: ExtendedAsyncSAEngine
    _config_provider: ManagerConfigProvider
    _valkey_stat_client: ValkeyStatClient
    _role_manager: RoleManager

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        config_provider: ManagerConfigProvider,
        valkey_stat_client: ValkeyStatClient,
    ) -> None:
        self._db = db
        self._config_provider = config_provider
        self._valkey_stat_client = valkey_stat_client
        self._role_manager = RoleManager()

    async def _get_group_by_id(self, session: SASession, group_id: uuid.UUID) -> Optional[GroupRow]:
        """Private method to get a group by ID using an existing session."""
        result = await session.execute(sa.select(GroupRow).where(groups.c.id == group_id))
        return result.scalar_one_or_none()

    @repository_decorator()
    async def create(self, creator: GroupCreator) -> GroupData:
        """Create a new group."""
        async with self._db.begin_session() as db_session:
            # Validate domain exists
            domain_exists = await db_session.scalar(
                sa.select(sa.exists().where(domains.c.name == creator.domain_name))
            )
            if not domain_exists:
                raise InvalidAPIParameters(
                    f"Cannot create group: Domain '{creator.domain_name}' does not exist"
                )

            # Validate resource policy exists
            policy_exists = await db_session.scalar(
                sa.select(
                    sa.exists().where(keypair_resource_policies.c.name == creator.resource_policy)
                )
            )
            if not policy_exists:
                raise InvalidAPIParameters(
                    f"Cannot create group: Resource policy '{creator.resource_policy}' does not exist"
                )

            # Check if group already exists
            check_stmt = sa.select(GroupRow).where(
                sa.and_(
                    GroupRow.name == creator.name,
                    GroupRow.domain_name == creator.domain_name,
                )
            )
            existing_group = await db_session.scalar(check_stmt)
            if existing_group is not None:
                raise InvalidAPIParameters(
                    f"Group with name '{creator.name}' already exists in domain '{creator.domain_name}'"
                )

            # Create the group
            row = GroupRow.from_creator(creator)
            db_session.add(row)
            await db_session.flush()
            await db_session.refresh(row)
            data = row.to_data()
            # Create RBAC role and permissions for the group
            await self._role_manager.create_system_role(db_session, data)

            return data

    @repository_decorator()
    async def modify_validated(
        self,
        group_id: uuid.UUID,
        modifier: GroupModifier,
        user_role: UserRole,
        user_update_mode: Optional[str] = None,
        user_uuids: Optional[list[uuid.UUID]] = None,
    ) -> Optional[GroupData]:
        """Modify a group with validation."""
        data = modifier.fields_to_update()

        if user_update_mode not in (None, "add", "remove"):
            raise ValueError("invalid user_update_mode")

        if not data and user_update_mode is None:
            return None

        async with self._db.begin_session() as session:
            # Handle user addition/removal
            if user_uuids and user_update_mode:
                if user_update_mode == "add":
                    values = [{"user_id": uuid, "group_id": group_id} for uuid in user_uuids]
                    await session.execute(
                        sa.insert(association_groups_users).values(values),
                    )
                elif user_update_mode == "remove":
                    await session.execute(
                        sa.delete(association_groups_users).where(
                            (association_groups_users.c.user_id.in_(user_uuids))
                            & (association_groups_users.c.group_id == group_id),
                        ),
                    )

            # Update group data if provided
            if data:
                update_stmt = (
                    sa.update(GroupRow)
                    .values(data)
                    .where(GroupRow.id == group_id)
                    .returning(GroupRow)
                )
                query_stmt = (
                    sa.select(GroupRow)
                    .from_statement(update_stmt)
                    .execution_options(populate_existing=True)
                )
                row = await session.scalar(query_stmt)
                row = cast(Optional[GroupRow], row)
                if row is None:
                    raise ProjectNotFound(f"Project not found: {group_id}")
                return row.to_data()

            # If only user updates were performed, return None
            return None

    @repository_decorator()
    async def mark_inactive(self, group_id: uuid.UUID) -> None:
        """Mark a group as inactive (soft delete)."""
        async with self._db.begin_session() as session:
            result = await session.execute(
                sa.update(groups)
                .values(
                    is_active=False,
                    integration_id=None,
                )
                .where(groups.c.id == group_id)
            )
            if result.rowcount > 0:
                return
            raise ProjectNotFound(f"Group not found: {group_id}")

    @repository_decorator()
    async def get_container_stats_for_period(
        self,
        start_date: datetime,
        end_date: datetime,
        group_ids: Optional[Sequence[UUID]] = None,
    ) -> list[dict]:
        """Get container statistics for groups within a time period."""
        async with self._db.begin_readonly() as conn:
            j = kernels.join(groups, groups.c.id == kernels.c.group_id).join(
                users, users.c.uuid == kernels.c.user_uuid
            )
            query = (
                sa.select([
                    kernels.c.id,
                    kernels.c.container_id,
                    kernels.c.session_id,
                    kernels.c.session_name,
                    kernels.c.access_key,
                    kernels.c.agent,
                    kernels.c.domain_name,
                    kernels.c.group_id,
                    kernels.c.attached_devices,
                    kernels.c.occupied_slots,
                    kernels.c.resource_opts,
                    kernels.c.vfolder_mounts,
                    kernels.c.mounts,
                    kernels.c.image,
                    kernels.c.status,
                    kernels.c.status_info,
                    kernels.c.status_changed,
                    kernels.c.last_stat,
                    kernels.c.status_history,
                    kernels.c.created_at,
                    kernels.c.terminated_at,
                    kernels.c.cluster_mode,
                    groups.c.name,
                    users.c.email,
                    users.c.full_name,
                ])
                .select_from(j)
                .where(
                    # Filter sessions which existence period overlaps with requested period
                    (
                        (kernels.c.terminated_at >= start_date)
                        & (kernels.c.created_at < end_date)
                        & (kernels.c.status.in_(RESOURCE_USAGE_KERNEL_STATUSES))
                    )
                    |
                    # Or, filter running sessions which created before requested end_date
                    ((kernels.c.created_at < end_date) & (kernels.c.status.in_(LIVE_STATUS))),
                )
                .order_by(sa.asc(kernels.c.terminated_at))
            )
            if group_ids:
                query = query.where(kernels.c.group_id.in_(group_ids))
            result = await conn.execute(query)
            rows = result.fetchall()

        kernel_ids = [str(row["id"]) for row in rows]
        raw_stats = await self._valkey_stat_client.get_user_kernel_statistics_batch(kernel_ids)

        objs_per_group = {}
        local_tz = self._config_provider.config.system.timezone

        for row, raw_stat in zip(rows, raw_stats):
            group_id = str(row["group_id"])
            last_stat = row["last_stat"]
            if not last_stat:
                if raw_stat is None:
                    log.warning("stat object for {} not found on redis, skipping", str(row["id"]))
                    continue
                last_stat = msgpack.unpackb(raw_stat)
            nfs = None
            if row["vfolder_mounts"]:
                # For >=22.03, return used host directories instead of volume host, which is not so useful.
                nfs = list(set([str(mount.host_path) for mount in row["vfolder_mounts"]]))
            elif row["mounts"] and isinstance(row["mounts"][0], list):
                # For the kernel records that have legacy contents of `mounts`.
                nfs = list(set([mount[2] for mount in row["mounts"]]))
            if row["terminated_at"] is None:
                used_time = used_days = None
            else:
                used_time = str(row["terminated_at"] - row["created_at"])
                used_days = (
                    row["terminated_at"].astimezone(local_tz).toordinal()
                    - row["created_at"].astimezone(local_tz).toordinal()
                    + 1
                )
            device_type = set()
            smp = 0
            gpu_mem_allocated = 0
            if row.attached_devices and row.attached_devices.get("cuda"):
                for dev_info in row.attached_devices["cuda"]:
                    if dev_info.get("model_name"):
                        device_type.add(dev_info["model_name"])
                    smp += int(nmget(dev_info, "data.smp", 0))
                    gpu_mem_allocated += int(nmget(dev_info, "data.mem", 0))
            gpu_allocated = 0
            if "cuda.devices" in row.occupied_slots:
                gpu_allocated = row.occupied_slots["cuda.devices"]
            if "cuda.shares" in row.occupied_slots:
                gpu_allocated = row.occupied_slots["cuda.shares"]
            c_info = {
                "id": str(row["id"]),
                "session_id": str(row["session_id"]),
                "container_id": row["container_id"],
                "domain_name": row["domain_name"],
                "group_id": str(row["group_id"]),
                "group_name": row["name"],
                "name": row["session_name"],
                "access_key": row["access_key"],
                "email": row["email"],
                "full_name": row["full_name"],
                "agent": row["agent"],
                "cpu_allocated": float(row.occupied_slots.get("cpu", 0)),
                "cpu_used": float(nmget(last_stat, "cpu_used.current", 0)),
                "mem_allocated": int(row.occupied_slots.get("mem", 0)),
                "mem_used": int(nmget(last_stat, "mem.capacity", 0)),
                "shared_memory": int(nmget(row.resource_opts, "shmem", 0)),
                "disk_allocated": 0,  # TODO: disk quota limit
                "disk_used": int(nmget(last_stat, "io_scratch_size/stats.max", 0, "/")),
                "io_read": int(nmget(last_stat, "io_read.current", 0)),
                "io_write": int(nmget(last_stat, "io_write.current", 0)),
                "used_time": used_time,
                "used_days": used_days,
                "device_type": list(device_type),
                "smp": float(smp),
                "gpu_mem_allocated": float(gpu_mem_allocated),
                "gpu_allocated": float(gpu_allocated),  # devices or shares
                "nfs": nfs,
                "image_id": row["image"],  # TODO: image id
                "image_name": row["image"],
                "created_at": str(row["created_at"]),
                "terminated_at": str(row["terminated_at"]),
                "status": row["status"].name,
                "status_info": row["status_info"],
                "status_changed": str(row["status_changed"]),
                "status_history": row["status_history"] or {},
                "cluster_mode": row["cluster_mode"],
            }
            if group_id not in objs_per_group:
                objs_per_group[group_id] = {
                    "domain_name": row["domain_name"],
                    "g_id": group_id,
                    "g_name": row["name"],  # this is group's name
                    "g_cpu_allocated": c_info["cpu_allocated"],
                    "g_cpu_used": c_info["cpu_used"],
                    "g_mem_allocated": c_info["mem_allocated"],
                    "g_mem_used": c_info["mem_used"],
                    "g_shared_memory": c_info["shared_memory"],
                    "g_disk_allocated": c_info["disk_allocated"],
                    "g_disk_used": c_info["disk_used"],
                    "g_io_read": c_info["io_read"],
                    "g_io_write": c_info["io_write"],
                    "g_device_type": copy.deepcopy(c_info["device_type"]),
                    "g_smp": c_info["smp"],
                    "g_gpu_mem_allocated": c_info["gpu_mem_allocated"],
                    "g_gpu_allocated": c_info["gpu_allocated"],
                    "c_infos": [c_info],
                }
            else:
                objs_per_group[group_id]["g_cpu_allocated"] += c_info["cpu_allocated"]
                objs_per_group[group_id]["g_cpu_used"] += c_info["cpu_used"]
                objs_per_group[group_id]["g_mem_allocated"] += c_info["mem_allocated"]
                objs_per_group[group_id]["g_mem_used"] += c_info["mem_used"]
                objs_per_group[group_id]["g_shared_memory"] += c_info["shared_memory"]
                objs_per_group[group_id]["g_disk_allocated"] += c_info["disk_allocated"]
                objs_per_group[group_id]["g_disk_used"] += c_info["disk_used"]
                objs_per_group[group_id]["g_io_read"] += c_info["io_read"]
                objs_per_group[group_id]["g_io_write"] += c_info["io_write"]
                for device in c_info["device_type"]:
                    if device not in objs_per_group[group_id]["g_device_type"]:
                        g_dev_type = objs_per_group[group_id]["g_device_type"]
                        g_dev_type.append(device)
                        objs_per_group[group_id]["g_device_type"] = list(set(g_dev_type))
                objs_per_group[group_id]["g_smp"] += c_info["smp"]
                objs_per_group[group_id]["g_gpu_mem_allocated"] += c_info["gpu_mem_allocated"]
                objs_per_group[group_id]["g_gpu_allocated"] += c_info["gpu_allocated"]
                objs_per_group[group_id]["c_infos"].append(c_info)
        return list(objs_per_group.values())

    @repository_decorator()
    async def fetch_project_resource_usage(
        self,
        start_date: datetime,
        end_date: datetime,
        project_ids: Optional[Sequence[UUID]] = None,
    ):
        """Fetch resource usage data for projects."""
        return await fetch_resource_usage(self._db, start_date, end_date, project_ids=project_ids)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact.types import ArtifactRevisionReadme
from ai.backend.manager.services.artifact_revision.actions.base import ArtifactRevisionAction


@dataclass
class GetArtifactRevisionReadmeAction(ArtifactRevisionAction):
    artifact_revision_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.artifact_revision_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_readme"


@dataclass
class GetArtifactRevisionReadmeActionResult(BaseActionResult):
    readme_data: ArtifactRevisionReadme

    @override
    def entity_id(self) -> Optional[str]:
        return None



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.artifact.repository import ArtifactRepository
from ai.backend.manager.repositories.image.repositories import RepositoryArgs


@dataclass
class ArtifactRepositories:
    repository: ArtifactRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ArtifactRepository(args.db)

        return cls(
            repository=repository,
        )



"""Factory functions for creating scheduler components."""

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.manager.clients.agent import AgentPool
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.plugin.network import NetworkPluginContext
from ai.backend.manager.repositories.deployment.repository import DeploymentRepository
from ai.backend.manager.repositories.scheduler import SchedulerRepository
from ai.backend.manager.sokovan.scheduler.allocators.repository_allocator import RepositoryAllocator
from ai.backend.manager.sokovan.scheduler.scheduler import (
    Scheduler,
    SchedulerArgs,
)
from ai.backend.manager.sokovan.scheduler.selectors.concentrated import ConcentratedAgentSelector
from ai.backend.manager.sokovan.scheduler.selectors.selector import AgentSelector
from ai.backend.manager.sokovan.scheduler.sequencers.fifo import FIFOSequencer
from ai.backend.manager.sokovan.scheduler.validators.concurrency import ConcurrencyValidator
from ai.backend.manager.sokovan.scheduler.validators.dependencies import DependenciesValidator
from ai.backend.manager.sokovan.scheduler.validators.domain_resource_limit import (
    DomainResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.group_resource_limit import (
    GroupResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.keypair_resource_limit import (
    KeypairResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.pending_session_count_limit import (
    PendingSessionCountLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.pending_session_resource_limit import (
    PendingSessionResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.reserved_batch import (
    ReservedBatchSessionValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.user_resource_limit import (
    UserResourceLimitValidator,
)
from ai.backend.manager.sokovan.scheduler.validators.validator import SchedulingValidator
from ai.backend.manager.types import DistributedLockFactory


def create_default_scheduler(
    repository: SchedulerRepository,
    deployment_repository: DeploymentRepository,
    config_provider: ManagerConfigProvider,
    lock_factory: DistributedLockFactory,
    agent_pool: AgentPool,
    network_plugin_ctx: NetworkPluginContext,
    event_producer: EventProducer,
    valkey_schedule: ValkeyScheduleClient,
) -> Scheduler:
    """
    Create a scheduler with default components.

    Args:
        repository: The repository for accessing system data
        deployment_repository: The deployment repository
        config_provider: The manager configuration provider
        lock_factory: Factory for creating distributed locks
        agent_pool: Pool for managing agent clients
        network_plugin_ctx: Network plugin context for network management
        event_producer: Event producer for publishing events
        valkey_schedule: Valkey client for scheduling operations

    Returns:
        A configured Scheduler instance
    """
    sequencer = FIFOSequencer()
    validator = SchedulingValidator([
        ConcurrencyValidator(),
        DependenciesValidator(),
        DomainResourceLimitValidator(),
        GroupResourceLimitValidator(),
        KeypairResourceLimitValidator(),
        PendingSessionCountLimitValidator(),
        PendingSessionResourceLimitValidator(),
        ReservedBatchSessionValidator(),
        UserResourceLimitValidator(),
    ])
    resource_priority = config_provider.config.manager.agent_selection_resource_priority
    agent_selector = AgentSelector(ConcentratedAgentSelector(resource_priority))
    allocator = RepositoryAllocator(repository)
    return Scheduler(
        SchedulerArgs(
            validator=validator,
            sequencer=sequencer,
            agent_selector=agent_selector,
            allocator=allocator,
            repository=repository,
            deployment_repository=deployment_repository,
            config_provider=config_provider,
            lock_factory=lock_factory,
            agent_pool=agent_pool,
            network_plugin_ctx=network_plugin_ctx,
            event_producer=event_producer,
            valkey_schedule=valkey_schedule,
        )
    )



"""empty message

Revision ID: 3cf19d906e71
Revises: 22964745c12b, 5d8e6043455e
Create Date: 2019-06-17 16:45:14.580560

"""

# revision identifiers, used by Alembic.
revision = "3cf19d906e71"
down_revision = ("22964745c12b", "5d8e6043455e")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.user_resource_policy.actions.create_user_resource_policy import (
    CreateUserResourcePolicyAction,
    CreateUserResourcePolicyActionResult,
)
from ai.backend.manager.services.user_resource_policy.actions.delete_user_resource_policy import (
    DeleteUserResourcePolicyAction,
    DeleteUserResourcePolicyActionResult,
)
from ai.backend.manager.services.user_resource_policy.actions.modify_user_resource_policy import (
    ModifyUserResourcePolicyAction,
    ModifyUserResourcePolicyActionResult,
)
from ai.backend.manager.services.user_resource_policy.service import UserResourcePolicyService


class UserResourcePolicyProcessors(AbstractProcessorPackage):
    create_user_resource_policy: ActionProcessor[
        CreateUserResourcePolicyAction, CreateUserResourcePolicyActionResult
    ]
    modify_user_resource_policy: ActionProcessor[
        ModifyUserResourcePolicyAction, ModifyUserResourcePolicyActionResult
    ]
    delete_user_resource_policy: ActionProcessor[
        DeleteUserResourcePolicyAction, DeleteUserResourcePolicyActionResult
    ]

    def __init__(
        self, service: UserResourcePolicyService, action_monitors: list[ActionMonitor]
    ) -> None:
        self.create_user_resource_policy = ActionProcessor(
            service.create_user_resource_policy, action_monitors
        )
        self.modify_user_resource_policy = ActionProcessor(
            service.modify_user_resource_policy, action_monitors
        )
        self.delete_user_resource_policy = ActionProcessor(
            service.delete_user_resource_policy, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateUserResourcePolicyAction.spec(),
            ModifyUserResourcePolicyAction.spec(),
            DeleteUserResourcePolicyAction.spec(),
        ]



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.services.resource_preset.actions.base import ResourcePresetAction
from ai.backend.manager.services.resource_preset.types import ResourcePresetModifier


@dataclass
class ModifyResourcePresetAction(ResourcePresetAction):
    modifier: ResourcePresetModifier
    id: Optional[uuid.UUID]
    name: Optional[str]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.id) if self.id else None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "modify"


@dataclass
class ModifyResourcePresetActionResult(BaseActionResult):
    resource_preset: ResourcePresetData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.resource_preset.id)


# TODO: Create exceptions.



from __future__ import annotations

import uuid
from collections.abc import Iterable, Sequence
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    Optional,
    Self,
    cast,
)

import graphene
import graphql
import more_itertools
import sqlalchemy as sa
import trafaret as t
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from sqlalchemy.engine.row import Row
from sqlalchemy.orm import joinedload, selectinload

from ai.backend.common import validators as tx
from ai.backend.common.defs.session import SESSION_PRIORITY_MAX, SESSION_PRIORITY_MIN
from ai.backend.common.exception import SessionWithInvalidStateError
from ai.backend.common.types import (
    ClusterMode,
    KernelId,
    ResourceSlot,
    SessionId,
    SessionResult,
    VFolderMount,
)
from ai.backend.manager.data.session.types import SessionData, SessionStatus
from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.idle import ReportInfo
from ai.backend.manager.models.gql_models.user import UserNode
from ai.backend.manager.models.kernel import KernelRow
from ai.backend.manager.models.utils import agg_to_array
from ai.backend.manager.services.session.actions.check_and_transit_status import (
    CheckAndTransitStatusAction,
)
from ai.backend.manager.services.session.actions.modify_session import (
    ModifySessionAction,
    SessionModifier,
)
from ai.backend.manager.types import OptionalState

from ..base import (
    BigInt,
    FilterExprArg,
    Item,
    OrderExprArg,
    PaginatedConnectionField,
    PaginatedList,
    batch_multiresult_in_session,
    batch_result_in_session,
    generate_sql_info_for_gql_connection,
)
from ..gql_relay import (
    AsyncNode,
    Connection,
    ConnectionResolverResult,
    GlobalIDField,
    ResolvedGlobalID,
)
from ..minilang import ArrayFieldItem, JSONFieldItem, ORMFieldItem
from ..minilang.ordering import ColumnMapType, QueryOrderParser
from ..minilang.queryfilter import FieldSpecType, QueryFilterParser
from ..rbac import ScopeType, SystemScope
from ..rbac.context import ClientContext
from ..rbac.permission_defs import ComputeSessionPermission
from ..rbac.permission_defs import VFolderPermission as VFolderRBACPermission
from ..session import (
    DEFAULT_SESSION_ORDERING,
    SessionDependencyRow,
    SessionRow,
    SessionTypes,
    by_domain_name,
    by_raw_filter,
    by_resource_group_name,
    by_status,
    get_permission_ctx,
)
from ..types import (
    QueryCondition,
    QueryOption,
    join_by_related_field,
    load_related_field,
)
from ..user import UserRole, UserRow
from ..vfolder import VFolderRow
from ..vfolder import get_permission_ctx as get_vfolder_permission_ctx
from .group import GroupRow
from .kernel import ComputeContainer, KernelConnection, KernelNode
from .vfolder import VirtualFolderConnection, VirtualFolderNode

if TYPE_CHECKING:
    from ..gql import GraphQueryContext

__all__ = (
    "ComputeSessionNode",
    "ComputeSessionConnection",
    "ModifyComputeSession",
    "ModifyComputeSession",
    "ComputeSession",
    "ComputeSessionList",
    "InferenceSession",
    "InferenceSessionList",
)


_queryfilter_fieldspec: FieldSpecType = {
    "id": ("id", None),
    "type": ("session_type", lambda s: SessionTypes(s)),
    "name": ("name", None),
    "priority": ("priority", None),
    "images": (ArrayFieldItem("images"), None),
    "image": (ArrayFieldItem("images"), None),
    "agent_ids": (ArrayFieldItem("agent_ids"), None),
    "domain_name": ("domain_name", None),
    "project_id": ("group_id", None),
    "user_id": ("user_uuid", None),
    "full_name": (ORMFieldItem(UserRow.full_name), None),
    "group_name": (ORMFieldItem(GroupRow.name), None),
    "user_email": (ORMFieldItem(UserRow.email), None),
    "access_key": ("access_key", None),
    "scaling_group": ("scaling_group_name", None),
    "cluster_mode": ("cluster_mode", lambda s: ClusterMode(s)),
    "cluster_size": ("cluster_size", None),
    "status": ("status", lambda s: SessionStatus(s)),
    "status_info": ("status_info", None),
    "result": ("result", lambda s: SessionResult(s)),
    "created_at": ("created_at", dtparse),
    "terminated_at": ("terminated_at", dtparse),
    "starts_at": ("starts_at", dtparse),
    "scheduled_at": (
        JSONFieldItem("status_history", SessionStatus.SCHEDULED.name),
        dtparse,
    ),
    "startup_command": ("startup_command", None),
}

_queryorder_colmap: ColumnMapType = {
    "id": ("id", None),
    "type": ("session_type", None),
    "name": ("name", None),
    "priority": ("priority", None),
    "images": ("images", None),
    "image": ("images", None),
    "agent_ids": ("agent_ids", None),
    "domain_name": ("domain_name", None),
    "project_id": ("group_id", None),
    "user_id": ("user_uuid", None),
    "access_key": ("access_key", None),
    "scaling_group": ("scaling_group_name", None),
    "cluster_mode": ("cluster_mode", None),
    "cluster_size": ("cluster_size", None),
    "status": ("status", None),
    "status_info": ("status_info", None),
    "result": ("result", None),
    "created_at": ("created_at", None),
    "terminated_at": ("terminated_at", None),
    "starts_at": ("starts_at", None),
    "scheduled_at": (
        JSONFieldItem("status_history", SessionStatus.SCHEDULED.name),
        None,
    ),
}


class SessionPermissionValueField(graphene.Scalar):
    class Meta:
        description = f"Added in 24.09.0. One of {[val.value for val in ComputeSessionPermission]}."

    @staticmethod
    def serialize(val: ComputeSessionPermission) -> str:
        return val.value

    @staticmethod
    def parse_literal(node: Any, _variables=None):
        if isinstance(node, graphql.language.ast.StringValueNode):
            return ComputeSessionPermission(node.value)

    @staticmethod
    def parse_value(value: str) -> ComputeSessionPermission:
        return ComputeSessionPermission(value)


class ComputeSessionNode(graphene.ObjectType):
    class Meta:
        interfaces = (AsyncNode,)
        description = "Added in 24.09.0."

    # identity
    row_id = graphene.UUID(description="ID of session.")
    tag = graphene.String()
    name = graphene.String()
    type = graphene.String()
    priority = graphene.Int(
        description="Added in 24.09.0.",
    )

    # cluster
    cluster_template = graphene.String()
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()

    # ownership
    domain_name = graphene.String()
    project_id = graphene.UUID()
    user_id = graphene.UUID()
    owner = graphene.Field(UserNode, description="Added in 25.13.0.")
    access_key = graphene.String()
    permissions = graphene.List(
        SessionPermissionValueField,
        description=f"One of {[val.value for val in ComputeSessionPermission]}.",
    )

    # status
    status = graphene.String()
    # status_changed = GQLDateTime()  # FIXME: generated attribute
    status_info = graphene.String()
    status_data = graphene.JSONString()
    status_history = graphene.JSONString()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    starts_at = GQLDateTime()
    scheduled_at = GQLDateTime()

    queue_position = graphene.Int(description="Added in 25.13.0.")

    startup_command = graphene.String()
    result = graphene.String()
    commit_status = graphene.String()
    abusing_reports = graphene.List(lambda: graphene.JSONString)
    idle_checks = graphene.JSONString()

    # resources
    agent_ids = graphene.List(lambda: graphene.String)
    resource_opts = graphene.JSONString()
    scaling_group = graphene.String()
    service_ports = graphene.JSONString()
    vfolder_mounts = graphene.List(lambda: graphene.String)
    occupied_slots = graphene.JSONString()
    requested_slots = graphene.JSONString()
    image_references = graphene.List(
        lambda: graphene.String,
        description="Added in 25.4.0.",
    )
    vfolder_nodes = PaginatedConnectionField(
        VirtualFolderConnection,
        description="Added in 25.4.0.",
    )

    # statistics
    num_queries = BigInt()
    inference_metrics = graphene.JSONString()

    # relations
    kernel_nodes = PaginatedConnectionField(
        KernelConnection,
    )
    dependents = PaginatedConnectionField(
        "ai.backend.manager.models.gql_models.session.ComputeSessionConnection",
        description="Added in 24.09.0.",
    )
    dependees = PaginatedConnectionField(
        "ai.backend.manager.models.gql_models.session.ComputeSessionConnection",
        description="Added in 24.09.0.",
    )
    graph = PaginatedConnectionField(
        "ai.backend.manager.models.gql_models.session.ComputeSessionConnection",
        description="Added in 24.09.0.",
    )

    @classmethod
    async def get_node(
        cls,
        info: graphene.ResolveInfo,
        id: str,
    ) -> Optional[Self]:
        graphene_ctx: GraphQueryContext = info.context
        _, raw_session_id = AsyncNode.resolve_global_id(info, id)
        async with graphene_ctx.db.begin_readonly_session() as db_session:
            stmt = (
                sa.select(SessionRow)
                .where(SessionRow.id == uuid.UUID(raw_session_id))
                .options(selectinload(SessionRow.kernels), joinedload(SessionRow.user))
            )
            query_result = await db_session.scalar(stmt)
            return cls.from_row(graphene_ctx, query_result) if query_result is not None else None

    @classmethod
    def _add_basic_options_to_query(
        cls, stmt: sa.sql.Select, is_count: bool = False
    ) -> sa.sql.Select:
        options = [
            join_by_related_field(SessionRow.user),
            join_by_related_field(SessionRow.group),
        ]
        if not is_count:
            options = [
                *options,
                load_related_field(SessionRow.kernel_load_option()),
                load_related_field(SessionRow.user_load_option(already_joined=True)),
                load_related_field(SessionRow.project_load_option(already_joined=True)),
            ]
        for option in options:
            stmt = option(stmt)
        return stmt

    async def resolve_queue_position(self, info: graphene.ResolveInfo) -> Optional[int]:
        if self.status != SessionStatus.PENDING:
            return None
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader_by_func(
            graph_ctx, self._batch_load_queue_position
        )
        return await loader.load(self.row_id)

    async def _batch_load_queue_position(
        self, ctx: GraphQueryContext, session_ids: Sequence[SessionId]
    ) -> list[Optional[int]]:
        positions = await ctx.valkey_schedule.get_queue_positions(session_ids)
        return positions

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: SessionRow,
        *,
        permissions: Optional[Iterable[ComputeSessionPermission]] = None,
    ) -> Self:
        status_history = row.status_history or {}
        raw_scheduled_at = status_history.get(SessionStatus.SCHEDULED.name)
        result = cls(
            # identity
            id=row.id,  # auto-converted to Relay global ID
            row_id=row.id,
            tag=row.tag,
            name=row.name,
            type=row.session_type,
            cluster_template=None,
            cluster_mode=row.cluster_mode,
            cluster_size=row.cluster_size,
            priority=row.priority,
            # ownership
            domain_name=row.domain_name,
            project_id=row.group_id,
            user_id=row.user_uuid,
            access_key=row.access_key,
            owner=UserNode.from_row(ctx, row.user),
            # status
            status=row.status.name,
            # status_changed=row.status_changed,  # FIXME: generated attribute
            status_info=row.status_info,
            status_data=row.status_data,
            status_history=status_history,
            created_at=row.created_at,
            starts_at=row.starts_at,
            terminated_at=row.terminated_at,
            scheduled_at=datetime.fromisoformat(raw_scheduled_at)
            if raw_scheduled_at is not None
            else None,
            startup_command=row.startup_command,
            result=row.result.name,
            # resources
            agent_ids=row.agent_ids,
            scaling_group=row.scaling_group_name,
            # TODO: Deprecate 'vfolder_mounts' and replace it with a list of VirtualFolderNodes
            vfolder_mounts=[vf.vfid.folder_id for vf in row.vfolders_sorted_by_id],
            occupied_slots=row.occupying_slots.to_json(),
            requested_slots=row.requested_slots.to_json(),
            image_references=row.images,
            service_ports=row.main_kernel.service_ports,
            # statistics
            num_queries=row.num_queries,
        )
        result.permissions = [] if permissions is None else permissions
        return result

    @classmethod
    def from_dataclass(
        cls,
        ctx: GraphQueryContext,
        session_data: SessionData,
        *,
        permissions: Optional[Iterable[ComputeSessionPermission]] = None,
    ) -> Self:
        status_history = session_data.status_history or {}
        raw_scheduled_at = status_history.get(SessionStatus.SCHEDULED.name)
        if not session_data.vfolder_mounts:
            vfolder_mounts = []
        else:
            vfolder_mounts = [vf.vfid.folder_id for vf in session_data.vfolder_mounts]

        if session_data.owner is None:
            raise SessionWithInvalidStateError()

        result = cls(
            # identity
            id=session_data.id,  # auto-converted to Relay global ID
            row_id=session_data.id,
            tag=session_data.tag,
            name=session_data.name,
            type=session_data.session_type,
            cluster_template=None,
            cluster_mode=session_data.cluster_mode,
            cluster_size=session_data.cluster_size,
            priority=session_data.priority,
            # ownership
            domain_name=session_data.domain_name,
            project_id=session_data.group_id,
            user_id=session_data.user_uuid,
            access_key=session_data.access_key,
            owner=UserNode.from_dataclass(ctx, session_data.owner),
            # status
            status=session_data.status.name,
            # status_changed=row.status_changed,  # FIXME: generated attribute
            status_info=session_data.status_info,
            status_data=session_data.status_data,
            status_history=status_history,
            created_at=session_data.created_at,
            starts_at=session_data.starts_at,
            terminated_at=session_data.terminated_at,
            scheduled_at=datetime.fromisoformat(raw_scheduled_at)
            if raw_scheduled_at is not None
            else None,
            startup_command=session_data.startup_command,
            result=session_data.result.name,
            # resources
            agent_ids=session_data.agent_ids,
            scaling_group=session_data.scaling_group_name,
            vfolder_mounts=vfolder_mounts,
            occupied_slots=session_data.occupying_slots,
            requested_slots=session_data.requested_slots,
            image_references=session_data.images,
            service_ports=session_data.service_ports,
            # statistics
            num_queries=session_data.num_queries,
        )
        result.permissions = [] if permissions is None else permissions
        return result

    async def resolve_idle_checks(self, info: graphene.ResolveInfo) -> dict[str, Any] | None:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader_by_func(
            graph_ctx, self.batch_load_idle_checks
        )
        return await loader.load(self.row_id)

    async def resolve_vfolder_nodes(
        self,
        info: graphene.ResolveInfo,
    ) -> ConnectionResolverResult[VirtualFolderNode]:
        ctx: GraphQueryContext = info.context
        _folder_ids = cast(list[uuid.UUID], self.vfolder_mounts)
        loader = ctx.dataloader_manager.get_loader_by_func(ctx, VirtualFolderNode.batch_load_by_id)
        result = cast(list[list[VirtualFolderNode]], await loader.load_many(_folder_ids))

        vf_nodes = cast(list[VirtualFolderNode], list(more_itertools.flatten(result)))

        # Calculate permissions for each node
        if vf_nodes:
            async with ctx.db.connect() as db_conn:
                user = ctx.user
                client_ctx = ClientContext(ctx.db, user["domain_name"], user["uuid"], user["role"])
                permission_ctx = await get_vfolder_permission_ctx(
                    db_conn, client_ctx, SystemScope(), VFolderRBACPermission.READ_ATTRIBUTE
                )

                # Load VFolderRow for each node to calculate permissions
                query = sa.select(VFolderRow).where(VFolderRow.id.in_(_folder_ids))
                async with ctx.db.begin_readonly_session(db_conn) as db_session:
                    vfolder_rows = {row.id: row for row in await db_session.scalars(query)}

                # Update permissions for each node
                for node in vf_nodes:
                    if node.row_id in vfolder_rows:
                        node.permissions = await permission_ctx.calculate_final_permission(
                            vfolder_rows[node.row_id]
                        )

        return ConnectionResolverResult(vf_nodes, None, None, None, total_count=len(vf_nodes))

    async def resolve_kernel_nodes(
        self,
        info: graphene.ResolveInfo,
    ) -> ConnectionResolverResult[KernelNode]:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, "KernelNode.by_session_id")
        kernel_nodes = await loader.load(self.row_id)
        return ConnectionResolverResult(
            kernel_nodes, None, None, None, total_count=len(kernel_nodes)
        )

    async def resolve_resource_opts(self, info: graphene.ResolveInfo) -> dict[str, Any]:
        ctx: GraphQueryContext = info.context
        loader = ctx.dataloader_manager.get_loader(ctx, "KernelNode.by_session_id")
        kernels = await loader.load(self.row_id)
        return {kernel.cluster_hostname: kernel.resource_opts for kernel in kernels}

    async def resolve_dependees(
        self,
        info: graphene.ResolveInfo,
    ) -> ConnectionResolverResult[Self]:
        ctx: GraphQueryContext = info.context
        # Get my dependees (myself is the dependent)
        loader = ctx.dataloader_manager.get_loader(ctx, "ComputeSessionNode.by_dependent_id")
        sessions = await loader.load(self.row_id)
        return ConnectionResolverResult(
            sessions,
            None,
            None,
            None,
            total_count=len(sessions),
        )

    async def resolve_dependents(
        self,
        info: graphene.ResolveInfo,
    ) -> ConnectionResolverResult[Self]:
        ctx: GraphQueryContext = info.context
        # Get my dependents (myself is the dependee)
        loader = ctx.dataloader_manager.get_loader(ctx, "ComputeSessionNode.by_dependee_id")
        sessions = await loader.load(self.row_id)
        return ConnectionResolverResult(
            sessions,
            None,
            None,
            None,
            total_count=len(sessions),
        )

    async def resolve_graph(
        self,
        info: graphene.ResolveInfo,
    ) -> ConnectionResolverResult[Self]:
        from ..session import SessionDependencyRow, SessionRow

        ctx: GraphQueryContext = info.context

        async with ctx.db.begin_readonly_session() as db_sess:
            dependency_cte = (
                sa.select(SessionRow.id)
                .filter(SessionRow.id == self.row_id)
                .cte(name="dependency_cte", recursive=True)
            )
            dependee = sa.select(SessionDependencyRow.depends_on).join(
                dependency_cte, SessionDependencyRow.session_id == dependency_cte.c.id
            )
            dependent = sa.select(SessionDependencyRow.session_id).join(
                dependency_cte, SessionDependencyRow.depends_on == dependency_cte.c.id
            )
            dependency_cte = dependency_cte.union_all(dependee).union_all(dependent)
            # Get the session IDs in the graph
            query = sa.select(dependency_cte.c.id)
            session_ids = (await db_sess.execute(query)).scalars().all()
            # Get the session rows in the graph
            query = (
                sa.select(SessionRow)
                .where(SessionRow.id.in_(session_ids))
                .options(
                    selectinload(SessionRow.user),
                )
            )
            session_rows = (await db_sess.execute(query)).scalars().all()

        # Convert into GraphQL node objects
        sessions = [type(self).from_row(ctx, r) for r in session_rows]
        return ConnectionResolverResult(
            sessions,
            None,
            None,
            None,
            total_count=len(sessions),
        )

    @classmethod
    async def batch_load_idle_checks(
        cls, ctx: GraphQueryContext, session_ids: Sequence[SessionId]
    ) -> list[dict[str, ReportInfo]]:
        check_result = await ctx.idle_checker_host.get_batch_idle_check_report(session_ids)
        return [check_result[sid] for sid in session_ids]

    @classmethod
    async def batch_load_by_dependee_id(
        cls, ctx: GraphQueryContext, session_ids: Sequence[SessionId]
    ) -> Sequence[Sequence[Self]]:
        from ..session import SessionDependencyRow, SessionRow

        async with ctx.db.begin_readonly_session() as db_sess:
            j = sa.join(
                SessionRow, SessionDependencyRow, SessionRow.id == SessionDependencyRow.depends_on
            )
            query = (
                sa.select(SessionRow)
                .select_from(j)
                .where(SessionDependencyRow.session_id.in_(session_ids))
            )
            return await batch_multiresult_in_session(
                ctx,
                db_sess,
                query,
                cls,
                session_ids,
                lambda row: row.id,
            )

    @classmethod
    async def batch_load_by_dependent_id(
        cls, ctx: GraphQueryContext, session_ids: Sequence[SessionId]
    ) -> Sequence[Sequence[Self]]:
        from ..session import SessionDependencyRow, SessionRow

        async with ctx.db.begin_readonly_session() as db_sess:
            j = sa.join(
                SessionRow, SessionDependencyRow, SessionRow.id == SessionDependencyRow.session_id
            )
            query = (
                sa.select(SessionRow)
                .select_from(j)
                .where(SessionDependencyRow.depends_on.in_(session_ids))
            )
            return await batch_multiresult_in_session(
                ctx,
                db_sess,
                query,
                cls,
                session_ids,
                lambda row: row.id,
            )

    @classmethod
    async def get_accessible_node(
        cls,
        info: graphene.ResolveInfo,
        id: ResolvedGlobalID,
        scope_id: ScopeType,
        permission: ComputeSessionPermission,
    ) -> Optional[Self]:
        graph_ctx: GraphQueryContext = info.context
        user = graph_ctx.user
        client_ctx = ClientContext(graph_ctx.db, user["domain_name"], user["uuid"], user["role"])
        _, session_id = id
        async with graph_ctx.db.connect() as db_conn:
            permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope_id, permission)
            cond = permission_ctx.query_condition
            if cond is None:
                return None
            query = sa.select(SessionRow).where(cond & (SessionRow.id == uuid.UUID(session_id)))
            query = cls._add_basic_options_to_query(query)
            async with graph_ctx.db.begin_readonly_session(db_conn) as db_session:
                session_row = await db_session.scalar(query)
        result = cls.from_row(
            graph_ctx,
            session_row,
            permissions=await permission_ctx.calculate_final_permission(session_row),
        )
        return result

    @classmethod
    async def get_accessible_connection(
        cls,
        info: graphene.ResolveInfo,
        scope_id: ScopeType,
        permission: ComputeSessionPermission,
        filter_expr: Optional[str] = None,
        order_expr: Optional[str] = None,
        offset: Optional[int] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        before: Optional[str] = None,
        last: Optional[int] = None,
    ) -> ConnectionResolverResult[Self]:
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(_queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(_queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            SessionRow,
            SessionRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            client_ctx = ClientContext(
                graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
            )
            permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope_id, permission)
            cond = permission_ctx.query_condition
            if cond is None:
                return ConnectionResolverResult([], cursor, pagination_order, page_size, 0)
            query = cls._add_basic_options_to_query(query.where(cond))
            cnt_query = cls._add_basic_options_to_query(cnt_query.where(cond), is_count=True)
            async with graph_ctx.db.begin_readonly_session(db_conn) as db_session:
                session_rows = (await db_session.scalars(query)).all()
                total_cnt = await db_session.scalar(cnt_query)
        result: list[Self] = [
            cls.from_row(
                graph_ctx,
                row,
                permissions=await permission_ctx.calculate_final_permission(row),
            )
            for row in session_rows
        ]
        return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)


class ComputeSessionConnection(Connection):
    class Meta:
        node = ComputeSessionNode
        description = "Added in 24.09.0."


class TotalResourceSlot(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)
        description = "Added in 25.5.0."

    occupied_slots = graphene.JSONString()
    requested_slots = graphene.JSONString()

    @classmethod
    async def get_data(
        cls,
        ctx: GraphQueryContext,
        statuses: Optional[Iterable[str]] = None,
        raw_filter: Optional[str] = None,
        domain_name: Optional[str] = None,
        resource_group_name: Optional[str] = None,
    ) -> Self:
        if statuses is not None:
            status_list = [SessionStatus[s] for s in statuses]
        else:
            status_list = None
        query_conditions: list[QueryCondition] = []
        if raw_filter is not None:
            query_conditions.append(by_raw_filter(_queryfilter_fieldspec, raw_filter))
        if status_list is not None:
            query_conditions.append(by_status(status_list))
        if domain_name is not None:
            query_conditions.append(by_domain_name(domain_name))
        if resource_group_name is not None:
            query_conditions.append(by_resource_group_name(resource_group_name))
        query_options: list[QueryOption] = [
            load_related_field(SessionRow.kernel_load_option()),
            join_by_related_field(SessionRow.user),
            join_by_related_field(SessionRow.group),
        ]
        session_rows = await SessionRow.list_session_by_condition(
            query_conditions, query_options, db=ctx.db
        )
        occupied_slots = ResourceSlot()
        requested_slots = ResourceSlot()
        for row in session_rows:
            occupied_slots += row.occupying_slots
            requested_slots += row.requested_slots
        occupied, requested = occupied_slots.to_json(), requested_slots.to_json()

        return TotalResourceSlot(
            occupied_slots=occupied,
            requested_slots=requested,
        )


def _validate_priority_input(priority: int) -> None:
    if not (SESSION_PRIORITY_MIN <= priority <= SESSION_PRIORITY_MAX):
        raise ValueError(
            f"The priority value {priority!r} is out of range: "
            f"[{SESSION_PRIORITY_MIN}, {SESSION_PRIORITY_MAX}]."
        )


def _validate_name_input(name: str) -> None:
    try:
        tx.SessionName().check(name)
    except t.DataError:
        raise ValueError(f"Not allowed session name (n:{name})")


class ModifyComputeSession(graphene.relay.ClientIDMutation):
    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)  # TODO: check if working

    class Meta:
        description = "Added in 24.09.0."

    class Input:
        id = GlobalIDField(required=True)
        name = graphene.String(required=False)
        priority = graphene.Int(required=False)
        client_mutation_id = graphene.String(required=False)  # automatic input from relay

    # Output fields
    item = graphene.Field(ComputeSessionNode)
    client_mutation_id = graphene.String()  # Relay output

    @classmethod
    async def mutate_and_get_payload(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        **input,
    ) -> ModifyComputeSession:
        graph_ctx: GraphQueryContext = info.context
        _, raw_session_id = cast(ResolvedGlobalID, input["id"])
        session_id = SessionId(uuid.UUID(raw_session_id))

        priority = input.get("priority", graphql.Undefined)
        if priority:
            _validate_priority_input(priority)

        name = input.get("name", graphql.Undefined)
        if name:
            _validate_name_input(name)

        result = await graph_ctx.processors.session.modify_session.wait_for_complete(
            ModifySessionAction(
                session_id=session_id,
                modifier=SessionModifier(
                    name=OptionalState[str].from_graphql(name),
                    priority=OptionalState[int].from_graphql(priority),
                ),
            )
        )

        return ModifyComputeSession(
            ComputeSessionNode.from_dataclass(graph_ctx, result.session_data),
            input.get("client_mutation_id"),
        )


class CheckAndTransitStatusInput(graphene.InputObjectType):
    class Meta:
        description = "Added in 24.12.0."

    ids = graphene.List(lambda: GlobalIDField, required=True)
    client_mutation_id = graphene.String(required=False)  # input for relay


class CheckAndTransitStatus(graphene.Mutation):
    allowed_roles = (UserRole.USER, UserRole.ADMIN, UserRole.SUPERADMIN)

    class Meta:
        description = "Added in 24.12.0"

    class Arguments:
        input = CheckAndTransitStatusInput(required=True)

    # Output fields
    item = graphene.List(lambda: ComputeSessionNode)
    client_mutation_id = graphene.String()  # Relay output

    @classmethod
    async def mutate(
        cls,
        root,
        info: graphene.ResolveInfo,
        input: CheckAndTransitStatusInput,
    ) -> CheckAndTransitStatus:
        graph_ctx: GraphQueryContext = info.context
        session_ids = [SessionId(sid) for _, sid in input.ids]

        user_role = cast(UserRole, graph_ctx.user["role"])
        user_id = cast(uuid.UUID, graph_ctx.user["uuid"])

        session_nodes = []
        for session_id in session_ids:
            action_result = (
                await graph_ctx.processors.session.check_and_transit_status.wait_for_complete(
                    CheckAndTransitStatusAction(
                        user_id=user_id,
                        user_role=user_role,
                        session_id=session_id,
                    )
                )
            )
            session_nodes.append(
                ComputeSessionNode.from_dataclass(graph_ctx, action_result.session_data)
            )

        return CheckAndTransitStatus(session_nodes, input.get("client_mutation_id"))


class ComputeSession(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    # identity
    session_id = graphene.UUID()  # identical to `id`
    main_kernel_id = graphene.UUID()
    tag = graphene.String()
    name = graphene.String()
    type = graphene.String()
    main_kernel_role = graphene.String()
    priority = graphene.Int(
        description="Added in 24.09.0.",
    )

    # image
    image = graphene.String()  # image for the main container
    architecture = graphene.String()  # image architecture for the main container
    registry = graphene.String()  # image registry for the main container
    cluster_template = graphene.String()
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()

    # ownership
    domain_name = graphene.String()
    group_name = graphene.String()
    group_id = graphene.UUID()
    user_email = graphene.String()
    full_name = graphene.String()
    user_id = graphene.UUID()
    access_key = graphene.String()
    created_user_email = graphene.String()
    created_user_id = graphene.UUID()

    # status
    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    status_data = graphene.JSONString()
    status_history = graphene.JSONString()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    starts_at = GQLDateTime()
    scheduled_at = GQLDateTime()
    startup_command = graphene.String()
    result = graphene.String()
    commit_status = graphene.String()
    abusing_reports = graphene.List(lambda: graphene.JSONString)
    idle_checks = graphene.JSONString()

    # resources
    agent_ids = graphene.List(lambda: graphene.String)
    agents = graphene.List(lambda: graphene.String)
    resource_opts = graphene.JSONString()
    scaling_group = graphene.String()
    service_ports = graphene.JSONString()
    mounts = graphene.List(lambda: graphene.String)
    vfolder_mounts = graphene.List(lambda: graphene.String)
    occupying_slots = graphene.JSONString()
    occupied_slots = graphene.JSONString()  # legacy
    requested_slots = graphene.JSONString(description="Added in 24.03.0.")

    # statistics
    num_queries = BigInt()

    # owned containers (aka kernels)
    containers = graphene.List(lambda: ComputeContainer)

    # relations
    dependencies = graphene.List(lambda: ComputeSession)

    inference_metrics = graphene.JSONString()

    @classmethod
    def parse_row(cls, ctx: GraphQueryContext, row: Row) -> Mapping[str, Any]:
        assert row is not None
        email = getattr(row, "email")
        full_name = getattr(row, "full_name")
        group_name = getattr(row, "group_name")
        row = row.SessionRow
        status_history = row.status_history or {}
        raw_scheduled_at = status_history.get(SessionStatus.SCHEDULED.name)
        # TODO: Deprecate 'mounts' and replace it with a list of VirtualFolderNodes
        mounts_set: set[str] = set()
        mounts: list[str] = []
        vfolder_mounts = cast(list[VFolderMount], row.vfolders_sorted_by_id)
        for mount in vfolder_mounts:
            if mount.name not in mounts_set:
                mounts.append(mount.name)
                mounts_set.add(mount.name)
        return {
            # identity
            "id": row.id,
            "session_id": row.id,
            "main_kernel_id": row.main_kernel.id,
            "tag": row.tag,
            "name": row.name,
            "type": row.session_type.name,
            "main_kernel_role": row.session_type.name,  # legacy
            "priority": row.priority,
            # image
            "image": row.images[0] if row.images is not None else "",
            "architecture": row.main_kernel.architecture,
            "registry": row.main_kernel.registry,
            "cluster_template": None,  # TODO: implement
            "cluster_mode": row.cluster_mode,
            "cluster_size": row.cluster_size,
            # ownership
            "domain_name": row.domain_name,
            "group_name": group_name[0],
            "group_id": row.group_id,
            "user_email": email,
            "full_name": full_name,
            "user_id": row.user_uuid,
            "access_key": row.access_key,
            "created_user_email": None,  # TODO: implement
            "created_user_id": None,  # TODO: implement
            # status
            "status": row.status.name,
            "status_changed": row.status_changed,
            "status_info": row.status_info,
            "status_data": row.status_data,
            "status_history": status_history,
            "created_at": row.created_at,
            "terminated_at": row.terminated_at,
            "starts_at": row.starts_at,
            "scheduled_at": (
                datetime.fromisoformat(raw_scheduled_at) if raw_scheduled_at is not None else None
            ),
            "startup_command": row.startup_command,
            "result": row.result.name,
            # resources
            "agent_ids": row.agent_ids,
            "agents": row.agent_ids,  # for backward compatibility
            "scaling_group": row.scaling_group_name,
            "service_ports": row.main_kernel.service_ports,
            "mounts": [*{mount.name for mount in vfolder_mounts}],
            # TODO: Deprecate 'vfolder_mounts' and replace it with a list of VirtualFolderNodes
            "vfolder_mounts": [*{vf.vfid.folder_id for vf in vfolder_mounts}],
            "occupying_slots": row.occupying_slots.to_json(),
            "occupied_slots": row.occupying_slots.to_json(),
            "requested_slots": row.requested_slots.to_json(),
            # statistics
            "num_queries": row.num_queries,
        }

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: Row | None) -> ComputeSession | None:
        if row is None:
            return None
        props = cls.parse_row(ctx, row)
        return cls(**props)

    async def resolve_inference_metrics(
        self, info: graphene.ResolveInfo
    ) -> Optional[Mapping[str, Any]]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(
            graph_ctx, "KernelStatistics.inference_metrics_by_kernel"
        )
        return await loader.load(self.id)

    async def resolve_containers(
        self,
        info: graphene.ResolveInfo,
    ) -> Iterable[ComputeContainer]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(graph_ctx, "ComputeContainer.by_session")
        return await loader.load(self.session_id)

    async def resolve_dependencies(
        self,
        info: graphene.ResolveInfo,
    ) -> Iterable[ComputeSession]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(graph_ctx, "ComputeSession.by_dependency")
        return await loader.load(self.id)

    async def resolve_commit_status(self, info: graphene.ResolveInfo) -> str:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(
            graph_ctx, "ComputeSession.commit_statuses"
        )
        return await loader.load(self.main_kernel_id)

    async def resolve_resource_opts(self, info: graphene.ResolveInfo) -> dict[str, Any]:
        containers = self.containers
        if containers is None:
            containers = await self.resolve_containers(info)
        if containers is None:
            return {}
        self.containers = containers
        return {cntr.cluster_hostname: cntr.resource_opts for cntr in containers}

    async def resolve_abusing_reports(
        self, info: graphene.ResolveInfo
    ) -> Iterable[Optional[Mapping[str, Any]]]:
        containers = self.containers
        if containers is None:
            containers = await self.resolve_containers(info)
        if containers is None:
            return []
        self.containers = containers
        return [(await con.resolve_abusing_report(info, self.access_key)) for con in containers]

    async def resolve_idle_checks(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        graph_ctx: GraphQueryContext = info.context
        return await graph_ctx.idle_checker_host.get_idle_check_report(self.session_id)

    _queryfilter_fieldspec: FieldSpecType = {
        "id": ("sessions_id", None),
        "type": ("sessions_session_type", SessionTypes),
        "name": ("sessions_name", None),
        "priority": ("sessions_priority", None),
        "image": (ArrayFieldItem("sessions_images"), None),
        "agent_ids": (ArrayFieldItem("sessions_agent_ids"), None),
        "agent_id": (ArrayFieldItem("sessions_agent_ids"), None),
        "agents": (ArrayFieldItem("sessions_agent_ids"), None),  # for backward compatibility
        "domain_name": ("sessions_domain_name", None),
        "group_name": ("group_name", None),
        "user_email": ("users_email", None),
        "user_id": ("sessions_user_uuid", None),
        "full_name": ("users_full_name", None),
        "access_key": ("sessions_access_key", None),
        "scaling_group": ("sessions_scaling_group_name", None),
        "cluster_mode": ("sessions_cluster_mode", ClusterMode),
        "cluster_size": ("sessions_cluster_size", None),
        "status": ("sessions_status", SessionStatus),
        "status_info": ("sessions_status_info", None),
        "result": ("sessions_result", SessionResult),
        "created_at": ("sessions_created_at", dtparse),
        "terminated_at": ("sessions_terminated_at", dtparse),
        "starts_at": ("sessions_starts_at", dtparse),
        "scheduled_at": (
            JSONFieldItem("sessions_status_history", SessionStatus.SCHEDULED.name),
            dtparse,
        ),
        "startup_command": ("sessions_startup_command", None),
    }

    _queryorder_colmap: ColumnMapType = {
        "id": ("sessions_id", None),
        "type": ("sessions_session_type", None),
        "name": ("sessions_name", None),
        "image": ("sessions_images", None),
        "priority": ("sessions_priority", None),
        "agent_ids": ("sessions_agent_ids", None),
        "agent_id": ("sessions_agent_ids", None),
        "agents": ("sessions_agent_ids", None),
        "domain_name": ("sessions_domain_name", None),
        "group_name": ("group_name", None),
        "user_email": ("users_email", None),
        "user_id": ("sessions_user_uuid", None),
        "full_name": ("users_full_name", None),
        "access_key": ("sessions_access_key", None),
        "scaling_group": ("sessions_scaling_group_name", None),
        "cluster_mode": ("sessions_cluster_mode", None),
        # "cluster_template": "cluster_template",
        "cluster_size": ("sessions_cluster_size", None),
        "status": ("sessions_status", None),
        "status_info": ("sessions_status_info", None),
        "result": ("sessions_result", None),
        "created_at": ("sessions_created_at", None),
        "terminated_at": ("sessions_terminated_at", None),
        "starts_at": ("sessions_starts_at", None),
        "scheduled_at": (
            JSONFieldItem("sessions_status_history", SessionStatus.SCHEDULED.name),
            None,
        ),
    }

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        access_key: Optional[str] = None,
        status: Optional[str] = None,
        filter: Optional[str] = None,
    ) -> int:
        if isinstance(status, str):
            status_list = [SessionStatus[s] for s in status.split(",")]
        elif isinstance(status, SessionStatus):
            status_list = [status]
        j = (
            # joins with GroupRow and UserRow do not need to be LEFT OUTER JOIN since those foreign keys are not nullable.
            sa.join(SessionRow, GroupRow, SessionRow.group_id == GroupRow.id)
            .join(UserRow, SessionRow.user_uuid == UserRow.uuid)
            .join(KernelRow, SessionRow.id == KernelRow.session_id)
        )
        query = sa.select([sa.func.count(sa.distinct(SessionRow.id))]).select_from(j)
        if domain_name is not None:
            query = query.where(SessionRow.domain_name == domain_name)
        if group_id is not None:
            query = query.where(SessionRow.group_id == group_id)
        if access_key is not None:
            query = query.where(SessionRow.access_key == access_key)
        if status is not None:
            query = query.where(SessionRow.status.in_(status_list))
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        access_key: Optional[str] = None,
        status: Optional[str] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[ComputeSession | None]:
        if status is None:
            status_list = None
        elif isinstance(status, str):
            status_list = [SessionStatus[s] for s in status.split(",")]
        elif isinstance(status, SessionStatus):
            status_list = [status]
        j = (
            # joins with GroupRow and UserRow do not need to be LEFT OUTER JOIN since those foreign keys are not nullable.
            sa.join(SessionRow, GroupRow, SessionRow.group_id == GroupRow.id).join(
                UserRow, SessionRow.user_uuid == UserRow.uuid
            )
        )
        query = (
            sa.select(
                SessionRow,
                agg_to_array(GroupRow.name).label("group_name"),
                UserRow.email,
                UserRow.full_name,
            )
            .select_from(j)
            .options(selectinload(SessionRow.kernels.and_(KernelRow.cluster_role == DEFAULT_ROLE)))
            .group_by(SessionRow, UserRow.email, UserRow.full_name)
            .limit(limit)
            .offset(offset)
        )
        if domain_name is not None:
            query = query.where(SessionRow.domain_name == domain_name)
        if group_id is not None:
            query = query.where(SessionRow.group_id == group_id)
        if access_key is not None:
            query = query.where(SessionRow.access_key == access_key)
        if status is not None:
            query = query.where(SessionRow.status.in_(status_list))
        if filter is not None:
            parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = parser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(*DEFAULT_SESSION_ORDERING)
        async with ctx.db.begin_readonly_session() as db_sess:
            return [cls.from_row(ctx, r) async for r in (await db_sess.stream(query))]

    @classmethod
    async def batch_load_detail(
        cls,
        ctx: GraphQueryContext,
        session_ids: Sequence[SessionId],
        *,
        domain_name: Optional[str] = None,
        access_key: Optional[str] = None,
    ) -> Sequence[ComputeSession | None]:
        j = sa.join(SessionRow, GroupRow, SessionRow.group_id == GroupRow.id).join(
            UserRow, SessionRow.user_uuid == UserRow.uuid
        )
        query = (
            sa.select(
                SessionRow,
                GroupRow.name.label("group_name"),
                UserRow.email,
                UserRow.full_name,
            )
            .select_from(j)
            .where(SessionRow.id.in_(session_ids))
            .options(selectinload(SessionRow.kernels))
        )
        if domain_name is not None:
            query = query.where(SessionRow.domain_name == domain_name)
        if access_key is not None:
            query = query.where(SessionRow.access_key == access_key)
        async with ctx.db.begin_readonly_session() as db_sess:
            return await batch_result_in_session(
                ctx,
                db_sess,
                query,
                cls,
                session_ids,
                lambda row: row.SessionRow.id,
            )

    @classmethod
    async def batch_load_by_dependency(
        cls,
        ctx: GraphQueryContext,
        session_ids: Sequence[SessionId],
    ) -> Sequence[Sequence[ComputeSession]]:
        j = sa.join(
            SessionRow,
            SessionDependencyRow,
            SessionRow.id == SessionDependencyRow.depends_on,
        )
        query = (
            sa.select(SessionRow)
            .select_from(j)
            .where(SessionDependencyRow.session_id.in_(session_ids))
            .options(selectinload(SessionRow.kernels))
        )
        async with ctx.db.begin_readonly_session() as db_sess:
            return await batch_multiresult_in_session(
                ctx,
                db_sess,
                query,
                cls,
                session_ids,
                lambda row: row.SessionRow.id,
            )

    @classmethod
    async def batch_load_commit_statuses(
        cls,
        ctx: GraphQueryContext,
        kernel_ids: Sequence[KernelId],
    ) -> Sequence[str]:
        commit_statuses = await ctx.registry.get_commit_status(kernel_ids)
        return [commit_statuses[kernel_id] for kernel_id in kernel_ids]


class ComputeSessionList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(ComputeSession, required=True)


class InferenceSession(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)


class InferenceSessionList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(InferenceSession, required=True)



# ruff: noqa: E402
from __future__ import annotations

import attrs

from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.services.processors import Processors


@attrs.define(auto_attribs=True, slots=True)
class StrawberryGQLContext:
    processors: Processors
    config_provider: ManagerConfigProvider



"""
Common definitions/constants used throughout the manager.
"""

import enum
import re
from typing import Final

from ai.backend.common.arch import CURRENT_ARCH
from ai.backend.common.arch import DEFAULT_IMAGE_ARCH as DEFAULT_IMAGE_ARCH_
from ai.backend.common.arch import arch_name_aliases as arch_name_aliases_
from ai.backend.common.types import SlotName, SlotTypes

INTRINSIC_SLOTS: Final = {
    SlotName("cpu"): SlotTypes("count"),
    SlotName("mem"): SlotTypes("bytes"),
}

INTRINSIC_SLOTS_MIN: Final = {
    # Values below are representing default value of image label -
    # so they must be string instead of Decimal.
    SlotName("cpu"): "1",  # 1 core
    SlotName("mem"): "1073741824",  # 1 GiB
}

arch_name_aliases: Final = arch_name_aliases_
DEFAULT_IMAGE_ARCH: Final = DEFAULT_IMAGE_ARCH_
MANAGER_ARCH: Final = CURRENT_ARCH

# The default chunk size used for streaming network traffic
DEFAULT_CHUNK_SIZE: Final = 16 * (2**20)  # 16 MiB

# The default container role name for multi-container sessions
DEFAULT_ROLE: Final = "main"

PASSWORD_PLACEHOLDER: Final = "*****"

_RESERVED_VFOLDER_PATTERNS = [r"^\.[a-z0-9]+rc$", r"^\.[a-z0-9]+_profile$"]
RESERVED_DOTFILES = [".terminfo", ".jupyter", ".ssh", ".ssh/authorized_keys", ".local", ".config"]
RESERVED_VFOLDERS = [
    ".terminfo",
    ".jupyter",
    ".tmux.conf",
    ".ssh",
    "/bin",
    "/boot",
    "/dev",
    "/etc",
    "/lib",
    "/lib64",
    "/media",
    "/mnt",
    "/opt",
    "/proc",
    "/root",
    "/run",
    "/sbin",
    "/srv",
    "/sys",
    "/tmp",
    "/usr",
    "/var",
    "/home",
]
RESERVED_VFOLDER_PATTERNS = [re.compile(x) for x in _RESERVED_VFOLDER_PATTERNS]

# Mapping between vfolder names and their in-container paths.
VFOLDER_DSTPATHS_MAP = {
    ".linuxbrew": "/home/linuxbrew/.linuxbrew",
}


# The unique identifiers for distributed locks.
# To be used with PostgreSQL advisory locks, the values are defined as integers.
class LockID(enum.IntEnum):
    LOCKID_TEST = 42
    LOCKID_SCHEDULE = 91
    LOCKID_CHECK_PRECOND = 92
    LOCKID_START = 93
    LOCKID_SCHEDULE_TIMER = 191
    LOCKID_CHECK_PRECOND_TIMER = 192
    LOCKID_START_TIMER = 198
    LOCKID_SCALE_TIMER = 193
    LOCKID_LOG_CLEANUP_TIMER = 195
    LOCKID_IDLE_CHECK_TIMER = 196
    LOCKID_SESSION_STATUS_UPDATE_TIMER = 197
    # Sokovan timers for each ScheduleType
    # Short cycle timers
    LOCKID_SOKOVAN_SCHEDULE_TIMER_SHORT = 201
    LOCKID_SOKOVAN_CHECK_PRECOND_TIMER_SHORT = 202
    LOCKID_SOKOVAN_START_TIMER_SHORT = 203
    LOCKID_SOKOVAN_TERMINATE_TIMER_SHORT = 205
    LOCKID_SOKOVAN_CHECK_PULLING_PROGRESS_TIMER_SHORT = 207
    LOCKID_SOKOVAN_CHECK_CREATING_PROGRESS_TIMER_SHORT = 208
    LOCKID_SOKOVAN_CHECK_TERMINATING_PROGRESS_TIMER_SHORT = 209
    # Long cycle timers
    LOCKID_SOKOVAN_SCHEDULE_TIMER_LONG = 211
    LOCKID_SOKOVAN_CHECK_PRECOND_TIMER_LONG = 212
    LOCKID_SOKOVAN_START_TIMER_LONG = 213
    LOCKID_SOKOVAN_TERMINATE_TIMER_LONG = 215
    LOCKID_SOKOVAN_SWEEP_TIMER = 216  # Sweep only has long timer
    LOCKID_SOKOVAN_CHECK_PULLING_PROGRESS_TIMER_LONG = 217
    LOCKID_SOKOVAN_CHECK_CREATING_PROGRESS_TIMER_LONG = 218
    LOCKID_SOKOVAN_CHECK_TERMINATING_PROGRESS_TIMER_LONG = 219
    # Retry timers (only long cycle - 30 seconds)
    LOCKID_SOKOVAN_RETRY_PREPARING_TIMER = 220
    LOCKID_SOKOVAN_RETRY_CREATING_TIMER = 221
    # Deployment locks
    LOCKID_DEPLOYMENT_AUTO_SCALER = 222  # Lock for deployment auto-scaler
    LOCKID_DEPLOYMENT_PROVISIONING_ROUTES = 223  # Lock for provisioning routes
    LOCKID_DEPLOYMENT_HEALTH_CHECK_ROUTES = 224  # Lock for health check routes
    LOCKID_DEPLOYMENT_RUNNING_ROUTES = 225  # Lock for running routes
    LOCKID_DEPLOYMENT_CHECK_PENDING = 226  # For operations checking PENDING sessions
    LOCKID_DEPLOYMENT_CHECK_REPLICA = 227  # For operations checking REPLICA sessions
    LOCKID_DEPLOYMENT_DESTROYING = 228  # For operations destroying deployments
    # Sokovan target status locks (prevent concurrent operations on same status)
    LOCKID_SOKOVAN_TARGET_PENDING = 230  # For operations targeting PENDING sessions
    LOCKID_SOKOVAN_TARGET_PREPARING = 231  # For operations targeting PREPARING/PULLING sessions
    LOCKID_SOKOVAN_TARGET_CREATING = 232  # For operations targeting CREATING/PREPARED sessions
    LOCKID_SOKOVAN_TARGET_TERMINATING = 233  # For operations targeting TERMINATING sessions


SERVICE_MAX_RETRIES = 5  # FIXME: make configurable

DEFAULT_KEYPAIR_RESOURCE_POLICY_NAME: Final = "default"
DEFAULT_KEYPAIR_RATE_LIMIT: Final = 10000

DEFAULT_SHARED_MEMORY_SIZE: Final[str] = "64m"
START_SESSION_TIMEOUT_SEC: Final[float] = 60 * 30  # 30 min


DEFAULT_METRIC_RANGE_VECTOR_TIMEWINDOW: Final[str] = "1m"

ARTIFACT_MAX_SCAN_LIMIT: Final[int] = 500



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.permission.role import (
    UserRoleAssignmentData,
    UserRoleAssignmentInput,
)
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class AssignRoleAction(RoleAction):
    input: UserRoleAssignmentInput

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.input.user_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "assign"


@dataclass
class AssignRoleActionResult(BaseActionResult):
    success: bool
    data: Optional[UserRoleAssignmentData]

    @override
    def entity_id(self) -> Optional[str]:
        return None
